"""async_upnp_client.const module."""

from dataclasses import dataclass
from datetime import date, datetime, time
from enum import Enum
from ipaddress import IPv4Address, IPv6Address
from typing import Any, Callable, Mapping, MutableMapping, NamedTuple
from xml.etree import ElementTree as ET

from async_upnp_client.utils import parse_date_time, require_tzinfo

IPvXAddress = IPv4Address | IPv6Address  # pylint: disable=invalid-name
AddressTupleV4Type = tuple[str, int]
AddressTupleV6Type = tuple[str, int, int, int]
AddressTupleVXType = AddressTupleV4Type | AddressTupleV6Type  # pylint: disable=invalid-name

NS = {
    "soap_envelope": "http://schemas.xmlsoap.org/soap/envelope/",
    "device": "urn:schemas-upnp-org:device-1-0",
    "service": "urn:schemas-upnp-org:service-1-0",
    "event": "urn:schemas-upnp-org:event-1-0",
    "control": "urn:schemas-upnp-org:control-1-0",
}


MIME_TO_UPNP_CLASS_MAPPING: Mapping[str, str] = {
    "audio": "object.item.audioItem",
    "video": "object.item.videoItem",
    "image": "object.item.imageItem",
    "application/dash+xml": "object.item.videoItem",
    "application/x-mpegurl": "object.item.videoItem",
    "application/vnd.apple.mpegurl": "object.item.videoItem",
}


STATE_VARIABLE_TYPE_MAPPING: Mapping[str, Mapping[str, Callable]] = {
    "ui1": {"type": int, "in": int, "out": str},
    "ui2": {"type": int, "in": int, "out": str},
    "ui4": {"type": int, "in": int, "out": str},
    "ui8": {"type": int, "in": int, "out": str},
    "i1": {"type": int, "in": int, "out": str},
    "i2": {"type": int, "in": int, "out": str},
    "i4": {"type": int, "in": int, "out": str},
    "i8": {"type": int, "in": int, "out": str},
    "int": {"type": int, "in": int, "out": str},
    "r4": {"type": float, "in": float, "out": str},
    "r8": {"type": float, "in": float, "out": str},
    "number": {"type": float, "in": float, "out": str},
    "fixed.14.4": {"type": float, "in": float, "out": str},
    "float": {"type": float, "in": float, "out": str},
    "char": {"type": str, "in": str, "out": str},
    "string": {"type": str, "in": str, "out": str},
    "boolean": {
        "type": bool,
        "in": lambda s: s.lower() in ["1", "true", "yes"],
        "out": lambda b: "1" if b else "0",
    },
    "bin.base64": {"type": str, "in": str, "out": str},
    "bin.hex": {"type": str, "in": str, "out": str},
    "uri": {"type": str, "in": str, "out": str},
    "uuid": {"type": str, "in": str, "out": str},
    "date": {"type": date, "in": parse_date_time, "out": lambda d: d.isoformat()},
    "dateTime": {
        "type": datetime,
        "in": parse_date_time,
        "out": lambda dt: dt.isoformat("T", "seconds"),
    },
    "dateTime.tz": {
        "type": datetime,
        "validator": require_tzinfo,
        "in": parse_date_time,
        "out": lambda dt: dt.isoformat("T", "seconds"),
    },
    "time": {
        "type": time,
        "in": parse_date_time,
        "out": lambda t: t.isoformat("seconds"),
    },
    "time.tz": {
        "type": time,
        "validator": require_tzinfo,
        "in": parse_date_time,
        "out": lambda t: t.isoformat("T", "seconds"),
    },
}


class DeviceIcon(NamedTuple):
    """Device icon."""

    mimetype: str
    width: int
    height: int
    depth: int
    url: str


class DeviceInfo(NamedTuple):
    """Device info."""

    device_type: str
    friendly_name: str
    manufacturer: str
    manufacturer_url: str | None
    model_description: str | None
    model_name: str
    model_number: str | None
    model_url: str | None
    serial_number: str | None
    udn: str
    upc: str | None
    presentation_url: str | None
    url: str
    icons: list[DeviceIcon]
    xml: ET.Element


class ServiceInfo(NamedTuple):
    """Service info."""

    service_id: str
    service_type: str
    control_url: str
    event_sub_url: str
    scpd_url: str
    xml: ET.Element


class ActionArgumentInfo(NamedTuple):
    """Action argument info."""

    name: str
    direction: str
    state_variable_name: str
    xml: ET.Element


class ActionInfo(NamedTuple):
    """Action info."""

    name: str
    arguments: list[ActionArgumentInfo]
    xml: ET.Element


@dataclass(frozen=True)
class HttpRequest:
    """HTTP request."""

    method: str
    url: str
    headers: Mapping[str, str]
    body: str | None


@dataclass(frozen=True)
class HttpResponse:
    """HTTP response."""

    status_code: int
    headers: Mapping[str, str]
    body: str | None


@dataclass(frozen=True)
class StateVariableTypeInfo:
    """State variable type info."""

    data_type: str
    data_type_mapping: Mapping[str, Callable]
    default_value: str | None
    allowed_value_range: Mapping[str, str | None]
    allowed_values: list[str] | None
    xml: ET.Element


@dataclass(frozen=True)
class EventableStateVariableTypeInfo(StateVariableTypeInfo):
    """Eventable State variable type info."""

    max_rate: float | None


@dataclass(frozen=True)
class StateVariableInfo:
    """State variable info."""

    name: str
    send_events: bool
    type_info: StateVariableTypeInfo
    xml: ET.Element


# Headers
SsdpHeaders = MutableMapping[str, Any]
NotificationType = str  # NT header
UniqueServiceName = str  # USN header
SearchTarget = str  # ST header
UniqueDeviceName = str  # UDN
DeviceOrServiceType = str


# Event handler
ServiceId = str  # SID


class NotificationSubType(str, Enum):
    """NTS header."""

    SSDP_ALIVE = "ssdp:alive"
    SSDP_BYEBYE = "ssdp:byebye"
    SSDP_UPDATE = "ssdp:update"


class SsdpSource(str, Enum):
    """SSDP source."""

    ADVERTISEMENT = "advertisement"
    SEARCH = "search"

    # More detailed
    SEARCH_ALIVE = "search_alive"
    SEARCH_CHANGED = "search_changed"

    # More detailed.
    ADVERTISEMENT_ALIVE = "advertisement_alive"
    ADVERTISEMENT_BYEBYE = "advertisement_byebye"
    ADVERTISEMENT_UPDATE = "advertisement_update"



"""Unit tests for profile."""

# pylint: disable=protected-access

import asyncio
import time
from datetime import timedelta
from unittest.mock import Mock

import pytest

from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import HttpResponse
from async_upnp_client.exceptions import (
    UpnpActionResponseError,
    UpnpCommunicationError,
    UpnpConnectionError,
    UpnpResponseError,
)
from async_upnp_client.profiles.dlna import DmrDevice
from async_upnp_client.profiles.igd import IgdDevice

from ..conftest import RESPONSE_MAP, UpnpTestNotifyServer, UpnpTestRequester, read_file


class TestUpnpProfileDevice:
    """Test UPnpProfileDevice."""

    @pytest.mark.asyncio
    async def test_action_exists(self) -> None:
        """Test getting existing action."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        # doesn't error
        assert profile._action("RC", "GetMute") is not None

    @pytest.mark.asyncio
    async def test_action_not_exists(self) -> None:
        """Test getting non-existing action."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        # doesn't error
        assert profile._action("RC", "NonExisting") is None

    @pytest.mark.asyncio
    async def test_icon(self) -> None:
        """Test getting an icon returns the best available."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        assert profile.icon == "http://dlna_dmr:1234/device_icon_120.png"

    @pytest.mark.asyncio
    async def test_is_profile_device(self) -> None:
        """Test is_profile_device works for root and embedded devices."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        embedded = await factory.async_create_device("http://dlna_dmr:1234/device_embedded.xml")
        no_services = await factory.async_create_device("http://dlna_dmr:1234/device_incomplete.xml")
        igd_device = await factory.async_create_device("http://igd:1234/device.xml")

        assert DmrDevice.is_profile_device(device) is True
        assert DmrDevice.is_profile_device(embedded) is True
        assert DmrDevice.is_profile_device(no_services) is False
        assert DmrDevice.is_profile_device(igd_device) is False

        assert IgdDevice.is_profile_device(device) is False
        assert IgdDevice.is_profile_device(embedded) is False
        assert IgdDevice.is_profile_device(no_services) is False
        assert IgdDevice.is_profile_device(igd_device) is True

    @pytest.mark.asyncio
    async def test_is_profile_device_non_strict(self) -> None:
        """Test is_profile_device works for root and embedded devices."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        embedded = await factory.async_create_device("http://dlna_dmr:1234/device_embedded.xml")
        no_services = await factory.async_create_device("http://dlna_dmr:1234/device_incomplete.xml")
        empty_descriptor = await factory.async_create_device("http://dlna_dmr:1234/device_with_empty_descriptor.xml")
        igd_device = await factory.async_create_device("http://igd:1234/device.xml")

        assert DmrDevice.is_profile_device(device) is True
        assert DmrDevice.is_profile_device(embedded) is True
        assert DmrDevice.is_profile_device(no_services) is False
        assert DmrDevice.is_profile_device(igd_device) is False

        assert IgdDevice.is_profile_device(device) is False
        assert IgdDevice.is_profile_device(embedded) is False
        assert IgdDevice.is_profile_device(no_services) is False
        assert IgdDevice.is_profile_device(empty_descriptor) is False
        assert IgdDevice.is_profile_device(igd_device) is True

    @pytest.mark.asyncio
    async def test_subscribe_manual_resubscribe(self) -> None:
        """Test subscribing, resub, unsub, without auto_resubscribe."""
        now = time.monotonic()
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        # Test subscription
        timeout = await profile.async_subscribe_services(auto_resubscribe=False)
        assert timeout is not None
        # Timeout incorporates time tolerance, and is minimal renewal time
        assert timedelta(seconds=149 - 60) <= timeout <= timedelta(seconds=151 - 60)

        assert set(profile._subscriptions.keys()) == {
            "uuid:dummy-avt1",
            "uuid:dummy-cm1",
            "uuid:dummy",
        }

        # 3 timeouts, ~ 150, ~ 175, and ~ 300 seconds
        timeouts = sorted(profile._subscriptions.values())
        assert timeouts[0] == pytest.approx(now + 150, abs=1)
        assert timeouts[1] == pytest.approx(now + 175, abs=1)
        assert timeouts[2] == pytest.approx(now + 300, abs=1)

        # Tweak timeouts to check resubscription did something
        entry = requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1")]
        requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1")] = HttpResponse(
            entry.status_code, {**entry.headers, "timeout": "Second-90"}, entry.body
        )

        # Check subscriptions again, now timeouts should have changed
        timeout = await profile.async_subscribe_services(auto_resubscribe=False)
        assert timeout is not None
        assert timedelta(seconds=89 - 60) <= timeout <= timedelta(seconds=91 - 60)
        assert set(profile._subscriptions.keys()) == {
            "uuid:dummy-avt1",
            "uuid:dummy-cm1",
            "uuid:dummy",
        }
        timeouts = sorted(profile._subscriptions.values())
        assert timeouts[0] == pytest.approx(now + 90, abs=1)
        assert timeouts[1] == pytest.approx(now + 150, abs=1)

        # Test unsubscription
        await profile.async_unsubscribe_services()
        assert not profile._subscriptions

    @pytest.mark.asyncio
    async def test_subscribe_auto_resubscribe(self) -> None:
        """Test subscribing, resub, unsub, with auto_resubscribe."""
        now = time.monotonic()
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        # Tweak timeouts to get a resubscription in a time suitable for testing.
        # Resubscription tolerance (60 seconds) + 1 second to get set up
        entry = requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1")]
        requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1")] = HttpResponse(
            entry.status_code,
            {
                **entry.headers,
                "timeout": "Second-61",
            },
            entry.body,
        )

        # Test subscription
        timeout = await profile.async_subscribe_services(auto_resubscribe=True)
        assert timeout is None
        assert profile.is_subscribed is True

        # Check subscriptions are correct
        assert set(profile._subscriptions.keys()) == {
            "uuid:dummy-avt1",
            "uuid:dummy-cm1",
            "uuid:dummy",
        }
        timeouts = sorted(profile._subscriptions.values())
        assert timeouts[0] == pytest.approx(now + 61, abs=1)
        assert timeouts[1] == pytest.approx(now + 150, abs=1)

        # Check task is running
        assert isinstance(profile._resubscriber_task, asyncio.Task)
        assert not profile._resubscriber_task.cancelled()
        assert not profile._resubscriber_task.done()

        # Re-tweak timeouts to check resubscription did something
        entry = requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/AVTransport1")]
        requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/AVTransport1")] = HttpResponse(
            entry.status_code,
            {
                **entry.headers,
                "timeout": "Second-90",
            },
            entry.body,
        )

        # Wait for an auto-resubscribe
        await asyncio.sleep(1.5)
        now = time.monotonic()

        # Check subscriptions and task again
        assert set(profile._subscriptions.keys()) == {
            "uuid:dummy-avt1",
            "uuid:dummy-cm1",
            "uuid:dummy",
        }
        timeouts = sorted(profile._subscriptions.values())
        assert timeouts[0] == pytest.approx(now + 61, abs=1)
        assert timeouts[1] == pytest.approx(now + 90, abs=1)
        assert isinstance(profile._resubscriber_task, asyncio.Task)
        assert not profile._resubscriber_task.cancelled()
        assert not profile._resubscriber_task.done()
        assert profile.is_subscribed is True

        # Unsubscribe
        await profile.async_unsubscribe_services()

        # Task and subscriptions should be gone
        assert profile._resubscriber_task is None
        assert not profile._subscriptions
        assert profile.is_subscribed is False

    @pytest.mark.asyncio
    async def test_subscribe_fail(self) -> None:
        """Test subscribing fails with UpnpError if device is offline."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        # First request is fine, 2nd raises an exception, when trying to subscribe
        requester.exceptions.append(None)
        requester.exceptions.append(UpnpCommunicationError())

        with pytest.raises(UpnpCommunicationError):
            await profile.async_subscribe_services(True)

        # Subscriptions and resubscribe task should not exist
        assert not profile._subscriptions
        assert profile._resubscriber_task is None
        assert profile.is_subscribed is False

    @pytest.mark.asyncio
    async def test_subscribe_rejected(self) -> None:
        """Test subscribing rejected by device."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        # All requests give a response error
        requester.exceptions.append(UpnpResponseError(status=501))
        requester.exceptions.append(UpnpResponseError(status=501))

        with pytest.raises(UpnpResponseError):
            await profile.async_subscribe_services(True)

        # Subscriptions and resubscribe task should not exist
        assert not profile._subscriptions
        assert profile._resubscriber_task is None
        assert profile.is_subscribed is False

    @pytest.mark.asyncio
    async def test_auto_resubscribe_fail(self) -> None:
        """Test auto-resubscription when the device goes offline."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)
        assert device.available is True

        # Register an event handler
        on_event_mock = Mock(return_value=None)
        profile.on_event = on_event_mock

        # Setup for auto-resubscription
        entry = requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1")]
        requester.response_map[("SUBSCRIBE", "http://dlna_dmr:1234/upnp/event/RenderingControl1")] = HttpResponse(
            entry.status_code, {**entry.headers, "timeout": "Second-61"}, entry.body
        )
        await profile.async_subscribe_services(auto_resubscribe=True)

        # Exception raised when trying to resubscribe and subsequent retry subscribe
        requester.exceptions.append(UpnpCommunicationError("resubscribe"))
        requester.exceptions.append(UpnpConnectionError("subscribe"))

        # Wait for an auto-resubscribe
        await asyncio.sleep(1.5)

        # Device should now be offline, and an event notification sent
        assert device.available is False
        on_event_mock.assert_called_once_with(device.services["urn:schemas-upnp-org:service:RenderingControl:1"], [])
        # Device will still be subscribed because a notification was sent via
        # on_event instead of raising an exception.
        assert profile.is_subscribed is True

        # Unsubscribe should still work
        await profile.async_unsubscribe_services()
        assert profile.is_subscribed is False

        # Task and subscriptions should be gone
        assert profile._resubscriber_task is None
        assert not profile._subscriptions
        assert profile.is_subscribed is False

    @pytest.mark.asyncio
    async def test_resubscribe_drops_stale_subscription(self) -> None:
        """Regression test for #302: stale subscriptions must be dropped, not spun on."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)

        on_event_mock = Mock(return_value=None)
        profile.on_event = on_event_mock

        await profile.async_subscribe_services(auto_resubscribe=False)
        assert profile.is_subscribed is True

        # Force every subscription's renewal_time well past the stale threshold.
        stale_time = time.monotonic() - 3600.0
        stale_sids = list(profile._subscriptions)
        for sid in stale_sids:
            profile._subscriptions[sid] = stale_time

        # Call the resubscribe path directly with notify_errors=True (as the loop does).
        # Without the fix this would leave the stale entries in place; with the fix every
        # stale entry is removed and listeners are notified with an empty state change.
        await profile._async_resubscribe_services(notify_errors=True)

        assert not profile._subscriptions

        notified_services = {call.args[0] for call in on_event_mock.call_args_list}
        expected_services = {event_handler.service_for_sid(sid) for sid in stale_sids}
        assert notified_services == expected_services
        for call in on_event_mock.call_args_list:
            assert call.args[1] == []

    @pytest.mark.asyncio
    async def test_subscribe_no_event_handler(self) -> None:
        """Test no event handler."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        profile = DmrDevice(device, event_handler=None)

        # Doesn't error, but also doesn't do anything.
        await profile.async_subscribe_services()

    @pytest.mark.asyncio
    async def test_poll_state_variables(self) -> None:
        """Test polling state variables by calling a Get* action."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        requester.response_map[("POST", "http://dlna_dmr:1234/upnp/control/AVTransport1")] = HttpResponse(
            200, {}, read_file("dlna/dmr/action_GetPositionInfo.xml")
        )

        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        notify_server = UpnpTestNotifyServer(
            requester=requester,
            source=("192.168.1.2", 8090),
        )
        event_handler = notify_server.event_handler
        profile = DmrDevice(device, event_handler=event_handler)
        assert device.available is True

        # Register an event handler, it should be called when variable is updated
        on_event_mock = Mock(return_value=None)
        profile.on_event = on_event_mock
        assert profile.is_subscribed is False

        # Check state variables are currently empty
        assert profile.media_track_number is None
        assert profile.media_duration is None
        assert profile.current_track_uri is None
        assert profile._current_track_meta_data is None
        assert profile.media_title is None
        assert profile.media_artist is None

        # Call the Get action
        await profile._async_poll_state_variables("AVT", ["GetPositionInfo"], InstanceID=0)

        # on_event should be called with all changed variables
        expected_service = device.services["urn:schemas-upnp-org:service:AVTransport:1"]
        expected_changes = [
            expected_service.state_variables[name]
            for name in (
                "CurrentTrack",
                "CurrentTrackDuration",
                "CurrentTrackMetaData",
                "CurrentTrackURI",
                "RelativeTimePosition",
                "AbsoluteTimePosition",
                "RelativeCounterPosition",
                "AbsoluteCounterPosition",
            )
        ]
        on_event_mock.assert_called_once_with(expected_service, expected_changes)

        # Corresponding state variables should be updated
        assert profile.media_track_number == 1
        assert profile.media_duration == 194
        assert profile.current_track_uri == "uri://1.mp3"
        assert profile._current_track_meta_data is not None
        assert profile.media_title == "Test track"
        assert profile.media_artist == "A & B > C"

    @pytest.mark.asyncio
    async def test_poll_state_variables_missing_action(self) -> None:
        """Test missing action used when polling state variables is handled gracefully."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        requester.response_map[("POST", "http://dlna_dmr:1234/upnp/control/AVTransport1")] = HttpResponse(
            200, {}, read_file("dlna/dmr/action_GetPositionInfo.xml")
        )

        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        profile = DmrDevice(device, event_handler=None)
        assert device.available is True

        # Register an event handler, it should be called when variable is updated
        on_event_mock = Mock(return_value=None)
        profile.on_event = on_event_mock
        assert profile.is_subscribed is False

        # Check state variables are currently empty
        assert profile.media_track_number is None
        assert profile.media_duration is None
        assert profile.current_track_uri is None
        assert profile._current_track_meta_data is None
        assert profile.media_title is None
        assert profile.media_artist is None

        # Call an invalid and a valid Get action, in one function call
        await profile._async_poll_state_variables("AVT", ["GetInvalidAction", "GetPositionInfo"], InstanceID=0)

        # Missing (invalid) action should have no effect on valid action

        # on_event should be called with all changed variables
        expected_service = device.services["urn:schemas-upnp-org:service:AVTransport:1"]
        expected_changes = [
            expected_service.state_variables[name]
            for name in (
                "CurrentTrack",
                "CurrentTrackDuration",
                "CurrentTrackMetaData",
                "CurrentTrackURI",
                "RelativeTimePosition",
                "AbsoluteTimePosition",
                "RelativeCounterPosition",
                "AbsoluteCounterPosition",
            )
        ]
        on_event_mock.assert_called_once_with(expected_service, expected_changes)

        # Corresponding state variables should be updated
        assert profile.media_track_number == 1
        assert profile.media_duration == 194
        assert profile.current_track_uri == "uri://1.mp3"
        assert profile._current_track_meta_data is not None
        assert profile.media_title == "Test track"
        assert profile.media_artist == "A & B > C"

    @pytest.mark.asyncio
    async def test_poll_state_variables_failed_action(self) -> None:
        """Test failed action used when polling state variables is handled gracefully."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        # Good action response
        requester.response_map[("POST", "http://dlna_dmr:1234/upnp/control/AVTransport1")] = HttpResponse(
            200, {}, read_file("dlna/dmr/action_GetPositionInfo.xml")
        )

        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        profile = DmrDevice(device, event_handler=None)
        assert device.available is True

        # Register an event handler, it should be called when variable is updated
        on_event_mock = Mock(return_value=None)
        profile.on_event = on_event_mock
        assert profile.is_subscribed is False

        # Check state variables are currently empty
        assert profile.media_track_number is None
        assert profile.media_duration is None
        assert profile.current_track_uri is None
        assert profile._current_track_meta_data is None
        assert profile.media_title is None
        assert profile.media_artist is None

        # Failed GetTransportInfo action resulting in an exception
        requester.exceptions.append(UpnpActionResponseError(status=500, error_code=602, error_desc="Not implemented"))

        # Call a failing and a valid Get action, in one function call
        await profile._async_poll_state_variables("AVT", ["GetTransportInfo", "GetPositionInfo"], InstanceID=0)

        # Missing (invalid) action should have no effect on valid action

        # on_event should be called with all changed variables
        expected_service = device.services["urn:schemas-upnp-org:service:AVTransport:1"]
        expected_changes = [
            expected_service.state_variables[name]
            for name in (
                "CurrentTrack",
                "CurrentTrackDuration",
                "CurrentTrackMetaData",
                "CurrentTrackURI",
                "RelativeTimePosition",
                "AbsoluteTimePosition",
                "RelativeCounterPosition",
                "AbsoluteCounterPosition",
            )
        ]
        on_event_mock.assert_called_once_with(expected_service, expected_changes)

        # Corresponding state variables should be updated
        assert profile.media_track_number == 1
        assert profile.media_duration == 194
        assert profile.current_track_uri == "uri://1.mp3"
        assert profile._current_track_meta_data is not None
        assert profile.media_title == "Test track"
        assert profile.media_artist == "A & B > C"



"""async_upnp_client.profiles.profile module."""

import asyncio
import logging
import time
from datetime import timedelta
from typing import Any, Sequence

from async_upnp_client.client import (
    EventCallbackType,
    UpnpAction,
    UpnpDevice,
    UpnpService,
    UpnpStateVariable,
    UpnpValueError,
)
from async_upnp_client.const import AddressTupleVXType
from async_upnp_client.event_handler import UpnpEventHandler
from async_upnp_client.exceptions import (
    UpnpConnectionError,
    UpnpError,
    UpnpResponseError,
)
from async_upnp_client.search import async_search
from async_upnp_client.ssdp import SSDP_MX
from async_upnp_client.utils import CaseInsensitiveDict

_LOGGER = logging.getLogger(__name__)


SUBSCRIBE_TIMEOUT = timedelta(minutes=9)
RESUBSCRIBE_TOLERANCE = timedelta(minutes=1)
RESUBSCRIBE_TOLERANCE_SECS = RESUBSCRIBE_TOLERANCE.total_seconds()


def find_device_of_type(device: UpnpDevice, device_types: list[str]) -> UpnpDevice:
    """Find the (embedded) UpnpDevice of any of the device types."""
    for device_ in device.all_devices:
        if device_.device_type in device_types:
            return device_

    raise UpnpError(f"Could not find device of type: {device_types}")


class UpnpProfileDevice:
    """
    Base class for UpnpProfileDevices.

    Override _SERVICE_TYPES for aliases. Override SERVICE_IDS for required
    service_id values.
    """

    DEVICE_TYPES: list[str] = []

    SERVICE_IDS: frozenset[str] = frozenset()

    _SERVICE_TYPES: dict[str, set[str]] = {}

    @classmethod
    async def async_search(
        cls, source: AddressTupleVXType | None = None, timeout: int = SSDP_MX
    ) -> set[CaseInsensitiveDict]:
        """
        Search for this device type.

        This only returns search info, not a profile itself.

        :param source_ip Source IP to scan from
        :param timeout Timeout to use
        :return: Set of devices (dicts) found
        """
        responses = set()

        async def on_response(data: CaseInsensitiveDict) -> None:
            if "st" in data and data["st"] in cls.DEVICE_TYPES:
                responses.add(data)

        await async_search(async_callback=on_response, source=source, timeout=timeout)

        return responses

    @classmethod
    async def async_discover(cls) -> set[CaseInsensitiveDict]:
        """Alias for async_search."""
        return await cls.async_search()

    @classmethod
    def is_profile_device(cls, device: UpnpDevice) -> bool:
        """Check for device's support of the profile defined in this (sub)class.

        The device must be (or have an embedded device) that matches the class
        device type, and it must provide all services that are defined by this
        class.
        """
        try:
            profile_device = find_device_of_type(device, cls.DEVICE_TYPES)
        except UpnpError:
            return False

        # Check that every service required by the subclass is declared by the device
        device_service_ids = {service.service_id for service in profile_device.all_services}
        if not cls.SERVICE_IDS.issubset(device_service_ids):
            return False

        return True

    def __init__(self, device: UpnpDevice, event_handler: UpnpEventHandler | None) -> None:
        """Initialize."""
        self.device = device
        self.profile_device = find_device_of_type(device, self.DEVICE_TYPES)
        self._event_handler = event_handler
        self.on_event: EventCallbackType | None = None
        self._icon: str | None = None
        # Map of SID to renewal timestamp (monotonic clock seconds)
        self._subscriptions: dict[str, float] = {}
        self._resubscriber_task: asyncio.Task | None = None

    @property
    def name(self) -> str:
        """Get the name of the device."""
        return self.profile_device.name

    @property
    def manufacturer(self) -> str:
        """Get the manufacturer of this device."""
        return self.profile_device.manufacturer

    @property
    def model_description(self) -> str | None:
        """Get the model description of this device."""
        return self.profile_device.model_description

    @property
    def model_name(self) -> str:
        """Get the model name of this device."""
        return self.profile_device.model_name

    @property
    def model_number(self) -> str | None:
        """Get the model number of this device."""
        return self.profile_device.model_number

    @property
    def serial_number(self) -> str | None:
        """Get the serial number of this device."""
        return self.profile_device.serial_number

    @property
    def udn(self) -> str:
        """Get the UDN of the device."""
        return self.profile_device.udn

    @property
    def device_type(self) -> str:
        """Get the device type of this device."""
        return self.profile_device.device_type

    @property
    def icon(self) -> str | None:
        """Get a URL for the biggest icon for this device."""
        if not self.profile_device.icons:
            return None

        if not self._icon:
            icon_mime_preference = {"image/png": 3, "image/jpeg": 2, "image/gif": 1}
            icons = [icon for icon in self.profile_device.icons if icon.url]
            icons = sorted(
                icons,
                # Sort by area, then colour depth, then preferred mimetype
                key=lambda icon: (
                    icon.width * icon.height,
                    icon.depth,
                    icon_mime_preference.get(icon.mimetype, 0),
                ),
                reverse=True,
            )
            self._icon = icons[0].url

        return self._icon

    def _service(self, service_type_abbreviation: str) -> UpnpService | None:
        """Get UpnpService by service_type or alias."""
        if not self.profile_device:
            return None

        if service_type_abbreviation not in self._SERVICE_TYPES:
            return None

        for service_type in self._SERVICE_TYPES[service_type_abbreviation]:
            service = self.profile_device.find_service(service_type)
            if service:
                return service

        return None

    def _state_variable(self, service_name: str, state_variable_name: str) -> UpnpStateVariable | None:
        """Get state_variable from service."""
        service = self._service(service_name)
        if not service:
            return None

        if not service.has_state_variable(state_variable_name):
            return None

        return service.state_variable(state_variable_name)

    def _action(self, service_name: str, action_name: str) -> UpnpAction | None:
        """Check if service has action."""
        service = self._service(service_name)
        if not service:
            return None

        if not service.has_action(action_name):
            return None

        return service.action(action_name)

    def _interesting_service(self, service: UpnpService) -> bool:
        """Check if service is a service we're interested in."""
        service_type = service.service_type
        for service_types in self._SERVICE_TYPES.values():
            if service_type in service_types:
                return True

        return False

    async def _async_resubscribe_services(self, now: float | None = None, notify_errors: bool = False) -> None:
        """Renew existing subscriptions.

        :param now: time.monotonic reference for current time
        :param notify_errors: Call on_event in case of error instead of raising
        """
        assert self._event_handler

        if now is None:
            now = time.monotonic()
        renewal_threshold = now - RESUBSCRIBE_TOLERANCE_SECS

        _LOGGER.debug("Resubscribing to services with threshold %f", renewal_threshold)

        for sid, renewal_time in list(self._subscriptions.items()):
            if renewal_time < renewal_threshold:
                _LOGGER.debug("Dropping stale subscription %s with renewal_time %f", sid, renewal_time)
                del self._subscriptions[sid]
                service = self._event_handler.service_for_sid(sid)
                if service and notify_errors:
                    self._on_event(service, [])
                continue

            _LOGGER.debug("Resubscribing to %s with renewal_time %f", sid, renewal_time)
            # Subscription is going to be changed, no matter what
            del self._subscriptions[sid]
            # Determine service for on_event call in case of failure
            service = self._event_handler.service_for_sid(sid)
            if not service:
                _LOGGER.error("Subscription for %s was lost", sid)
                continue

            try:
                new_sid, timeout = await self._event_handler.async_resubscribe(sid, timeout=SUBSCRIBE_TIMEOUT)
            except UpnpError as err:
                if isinstance(err, UpnpConnectionError):
                    # Device has gone offline
                    self.profile_device.available = False
                _LOGGER.warning("Failed (re-)subscribing to: %s, reason: %r", sid, err)
                if notify_errors:
                    # Notify event listeners that something has changed
                    self._on_event(service, [])
                else:
                    raise
            else:
                self._subscriptions[new_sid] = now + timeout.total_seconds()

    async def _resubscribe_loop(self) -> None:
        """Periodically resubscribes to current subscriptions."""
        _LOGGER.debug("_resubscribe_loop started")
        while self._subscriptions:
            next_renewal = min(self._subscriptions.values())
            wait_time = next_renewal - time.monotonic() - RESUBSCRIBE_TOLERANCE_SECS
            _LOGGER.debug("Resubscribing in %f seconds", wait_time)
            if wait_time > 0:
                await asyncio.sleep(wait_time)

            await self._async_resubscribe_services(notify_errors=True)

        _LOGGER.debug("_resubscribe_loop ended because of no subscriptions")

    async def _update_resubscriber_task(self) -> None:
        """Start or stop the resubscriber task, depending on having subscriptions."""
        # Clear out done task to make later logic easier
        if self._resubscriber_task and self._resubscriber_task.cancelled():
            self._resubscriber_task = None

        if self._subscriptions and not self._resubscriber_task:
            _LOGGER.debug("Creating resubscribe_task")
            # pylint: disable=fixme
            self._resubscriber_task = asyncio.create_task(
                self._resubscribe_loop(),
                name=f"UpnpProfileDevice({self.name})._resubscriber_task",
            )

        if not self._subscriptions and self._resubscriber_task:
            _LOGGER.debug("Cancelling resubscribe_task")
            self._resubscriber_task.cancel()
            try:
                await self._resubscriber_task
            except asyncio.CancelledError:
                pass
            self._resubscriber_task = None

    async def async_subscribe_services(self, auto_resubscribe: bool = False) -> timedelta | None:
        """(Re-)Subscribe to services.

        :param auto_resubscribe: Automatically resubscribe to subscriptions
            before they expire. If this is enabled, failure to resubscribe will
            be indicated by on_event being called with the failed service and an
            empty state_variables list.
        :return: time until this next needs to be called, or None if manual
            resubscription is not needed.
        :raise UpnpResponseError: Device rejected subscription request.
            State variables will need to be polled.
        :raise UpnpError or subclass: Failed to subscribe to all interesting
            services.
        """
        if not self._event_handler:
            _LOGGER.debug("No event_handler, event handling disabled")
            return None

        # Using time.monotonic to avoid problems with system clock changes
        now = time.monotonic()

        try:
            if self._subscriptions:
                # Resubscribe existing subscriptions
                await self._async_resubscribe_services(now)
            else:
                # Subscribe to services we are interested in
                for service in self.profile_device.all_services:
                    if not self._interesting_service(service):
                        continue

                    _LOGGER.debug("Subscribing to service: %s", service)
                    service.on_event = self._on_event
                    new_sid, timeout = await self._event_handler.async_subscribe(service, timeout=SUBSCRIBE_TIMEOUT)
                    self._subscriptions[new_sid] = now + timeout.total_seconds()
        except UpnpError as err:
            if isinstance(err, UpnpResponseError) and not self._subscriptions:
                _LOGGER.debug("Device rejected subscription request: %r", err)
            else:
                _LOGGER.warning("Failed subscribing to service: %r", err)
            # Unsubscribe anything that was subscribed, no half-done subscriptions
            try:
                await self.async_unsubscribe_services()
            except UpnpError:
                pass
            raise

        if not self._subscriptions:
            return None

        if auto_resubscribe:
            await self._update_resubscriber_task()
            return None

        lowest_timeout_delta = min(self._subscriptions.values()) - now
        resubcription_timeout = timedelta(seconds=lowest_timeout_delta) - RESUBSCRIBE_TOLERANCE
        return max(resubcription_timeout, timedelta(seconds=0))

    async def _async_unsubscribe_service(self, sid: str) -> None:
        """Unsubscribe from one service, handling possible exceptions."""
        assert self._event_handler

        try:
            await self._event_handler.async_unsubscribe(sid)
        except UpnpError as err:
            _LOGGER.debug("Failed unsubscribing from: %s, reason: %r", sid, err)
        except KeyError:
            _LOGGER.warning(
                "%s was already unsubscribed. AiohttpNotifyServer was probably stopped before we could unsubscribe.",
                sid,
            )

    async def async_unsubscribe_services(self) -> None:
        """Unsubscribe from all of our subscribed services."""
        # Delete list of subscriptions and cancel renewal before unsubscribing
        # to avoid unsub-resub race.
        sids = list(self._subscriptions)
        self._subscriptions.clear()
        await self._update_resubscriber_task()

        await asyncio.gather(*(self._async_unsubscribe_service(sid) for sid in sids))

    @property
    def is_subscribed(self) -> bool:
        """Get current service subscription state."""
        return bool(self._subscriptions)

    async def _async_poll_state_variables(
        self, service_name: str, action_names: str | Sequence[str], **in_args: Any
    ) -> None:
        """Update state variables by polling actions that return their values.

        Assumes that the actions's relatedStateVariable names the correct state
        variable for updating.
        """
        service = self._service(service_name)
        if not service:
            _LOGGER.debug("Can't poll missing service %s", service_name)
            return

        if isinstance(action_names, str):
            action_names = [action_names]

        changed_state_variables: list[UpnpStateVariable] = []

        for action_name in action_names:
            try:
                action = service.action(action_name)
            except KeyError:
                _LOGGER.debug(
                    "Can't poll missing action %s:%s for state variables",
                    service_name,
                    action_name,
                )
                continue
            try:
                result = await action.async_call(**in_args)
            except UpnpResponseError as err:
                _LOGGER.debug(
                    "Failed to call action %s:%s for state variables: %r",
                    service_name,
                    action_name,
                    err,
                )
                continue

            for arg in action.arguments:
                if arg.direction != "out":
                    continue
                if arg.name not in result:
                    continue
                if arg.related_state_variable.value == arg.value:
                    continue

                try:
                    arg.related_state_variable.value = arg.value
                except UpnpValueError:
                    continue
                changed_state_variables.append(arg.related_state_variable)

        if changed_state_variables:
            self._on_event(service, changed_state_variables)

    def _on_event(self, service: UpnpService, state_variables: Sequence[UpnpStateVariable]) -> None:
        """
        State variable(s) changed. Override to handle events.

        :param service Service which sent the event.
        :param state_variables State variables which have been changed.
        """
        if self.on_event:
            self.on_event(service, state_variables)  # pylint: disable=not-callable



"""Unit tests for client_factory and client modules."""

from datetime import datetime, timedelta, timezone
from typing import MutableMapping

import defusedxml.ElementTree as DET
import pytest

from async_upnp_client.client import UpnpStateVariable
from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import HttpResponse
from async_upnp_client.exceptions import (
    UpnpActionError,
    UpnpActionErrorCode,
    UpnpActionResponseError,
    UpnpError,
    UpnpResponseError,
    UpnpValueError,
    UpnpXmlContentError,
    UpnpXmlParseError,
)

from .conftest import RESPONSE_MAP, UpnpTestRequester, read_file


class TestUpnpStateVariable:
    """Tests for UpnpStateVariable."""

    @pytest.mark.asyncio
    async def test_init(self) -> None:
        """Test initialization of a UpnpDevice."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        assert device
        assert device.device_type == "urn:schemas-upnp-org:device:MediaRenderer:1"

        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        assert service

        service_by_id = device.service_id("urn:upnp-org:serviceId:RenderingControl")
        assert service_by_id == service

        state_var = service.state_variable("Volume")
        assert state_var

        action = service.action("GetVolume")
        assert action

        argument = action.argument("InstanceID")
        assert argument

    @pytest.mark.asyncio
    async def test_init_embedded_device(self) -> None:
        """Test initialization of a embedded UpnpDevice."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://igd:1234/device.xml")
        assert device
        assert device.device_type == "urn:schemas-upnp-org:device:InternetGatewayDevice:1"

        embedded_device = device.embedded_devices["urn:schemas-upnp-org:device:WANDevice:1"]
        assert embedded_device
        assert embedded_device.device_type == "urn:schemas-upnp-org:device:WANDevice:1"
        assert embedded_device.parent_device == device

    @pytest.mark.asyncio
    async def test_init_xml(self) -> None:
        """Test XML is stored on every part of the UpnpDevice."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        assert device.xml is not None

        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        assert service.xml is not None

        state_var = service.state_variable("Volume")
        assert state_var.xml is not None

        action = service.action("GetVolume")
        assert action.xml is not None

        argument = action.argument("InstanceID")
        assert argument is not None
        assert argument.xml is not None

    @pytest.mark.asyncio
    async def test_init_bad_xml(self) -> None:
        """Test missing device element in device description."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/device.xml")] = HttpResponse(
            200,
            {},
            read_file("dlna/dmr/device_bad_namespace.xml"),
        )
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        with pytest.raises(UpnpXmlContentError):
            await factory.async_create_device("http://dlna_dmr:1234/device.xml")

    @pytest.mark.asyncio
    async def test_empty_descriptor(self) -> None:
        """Test device with an empty descriptor file called in description.xml."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/device.xml")] = HttpResponse(
            200,
            {},
            read_file("dlna/dmr/device_with_empty_descriptor.xml"),
        )
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        with pytest.raises(UpnpXmlParseError):
            await factory.async_create_device("http://dlna_dmr:1234/device.xml")

    @pytest.mark.asyncio
    async def test_empty_descriptor_non_strict(self) -> None:
        """Test device with an empty descriptor file called in description.xml."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/device.xml")] = HttpResponse(
            200,
            {},
            read_file("dlna/dmr/device_with_empty_descriptor.xml"),
        )
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester, non_strict=True)
        await factory.async_create_device("http://dlna_dmr:1234/device.xml")

    @pytest.mark.asyncio
    async def test_set_value_volume(self) -> None:
        """Test calling parsing/reading values from UpnpStateVariable."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("Volume")

        state_var.value = 10
        assert state_var.value == 10
        assert state_var.upnp_value == "10"

        state_var.upnp_value = "20"
        assert state_var.value == 20
        assert state_var.upnp_value == "20"

    @pytest.mark.asyncio
    async def test_set_value_mute(self) -> None:
        """Test setting a boolean value."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("Mute")

        state_var.value = True
        assert state_var.value is True
        assert state_var.upnp_value == "1"

        state_var.value = False
        assert state_var.value is False
        assert state_var.upnp_value == "0"

        state_var.upnp_value = "1"
        assert state_var.value is True
        assert state_var.upnp_value == "1"

        state_var.upnp_value = "0"
        assert state_var.value is False
        assert state_var.upnp_value == "0"

    @pytest.mark.asyncio
    async def test_value_min_max(self) -> None:
        """Test min/max restrictions."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("Volume")

        assert state_var.min_value == 0
        assert state_var.max_value == 100
        assert state_var.step_value == 1

        state_var.value = 10
        assert state_var.value == 10

        try:
            state_var.value = -10
            assert False
        except UpnpValueError:
            pass

        try:
            state_var.value = 110
            assert False
        except UpnpValueError:
            pass

    @pytest.mark.asyncio
    async def test_value_min_max_validation_disable(self) -> None:
        """Test if min/max validations can be disabled."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("Volume")

        # min/max/step are set
        assert state_var.min_value == 0
        assert state_var.max_value == 100
        assert state_var.step_value == 1

        # min/max are not validated
        state_var.value = -10
        assert state_var.value == -10

        state_var.value = 110
        assert state_var.value == 110

    @pytest.mark.asyncio
    async def test_value_allowed_value(self) -> None:
        """Test handling allowed values."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("A_ARG_TYPE_Channel")

        assert state_var.allowed_values == {"Master"}
        assert state_var.normalized_allowed_values == {"master"}

        # should be ok
        state_var.value = "Master"
        assert state_var.value == "Master"

        try:
            state_var.value = "Left"
            assert False
        except UpnpValueError:
            pass

    @pytest.mark.asyncio
    async def test_value_upnp_value_error(self) -> None:
        """Test handling invalid values in response."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("Volume")

        # should be ok
        state_var.upnp_value = "50"
        assert state_var.value == 50

        # should set UpnpStateVariable.UPNP_VALUE_ERROR
        state_var.upnp_value = "abc"
        assert state_var.value is None
        assert state_var.value_unchecked is UpnpStateVariable.UPNP_VALUE_ERROR

    @pytest.mark.asyncio
    async def test_value_date_time(self) -> None:
        """Test parsing of datetime."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("SV1")

        # should be ok
        state_var.upnp_value = "1985-04-12T10:15:30"
        assert state_var.value == datetime(1985, 4, 12, 10, 15, 30)

    @pytest.mark.asyncio
    async def test_value_date_time_tz(self) -> None:
        """Test parsing of date_time with a timezone."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("SV2")
        assert state_var is not None

        # should be ok
        state_var.upnp_value = "1985-04-12T10:15:30+0400"
        assert state_var.value == datetime(1985, 4, 12, 10, 15, 30, tzinfo=timezone(timedelta(hours=4)))
        assert state_var.value.tzinfo is not None

    @pytest.mark.asyncio
    async def test_send_events(self) -> None:
        """Test if send_events is properly handled."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")

        state_var = service.state_variable("A_ARG_TYPE_InstanceID")  # old style
        assert state_var.send_events is False

        state_var = service.state_variable("A_ARG_TYPE_Channel")  # new style
        assert state_var.send_events is False

        state_var = service.state_variable("Volume")  # broken/none given
        assert state_var.send_events is False

        state_var = service.state_variable("LastChange")
        assert state_var.send_events is True

    @pytest.mark.asyncio
    async def test_big_ints(self) -> None:
        """Test state variable types i8 and ui8."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dms:1234/ContentDirectory_1.xml")] = HttpResponse(
            200,
            {},
            read_file("scpd_i8.xml"),
        )
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dms:1234/device.xml")
        assert device is not None

    @pytest.mark.asyncio
    async def test_parse_state_variable_datetime_invalid_case_non_strict(self) -> None:
        """Test that non-standard 'DateTime' is accepted in non-strict mode (issue #299)."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/RenderingControl_1.xml")] = HttpResponse(
            200,
            {},
            read_file("dlna/dmr/RenderingControl_1_invalid_datetime_case.xml"),
        )

        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        state_var = service.state_variable("SV1")

        state_var.upnp_value = "1985-04-12T10:15:30"
        assert state_var.value == datetime(1985, 4, 12, 10, 15, 30)

    @pytest.mark.asyncio
    async def test_parse_state_variable_datetime_invalid_case_strict(self) -> None:
        """Test that non-standard 'DateTime' fails in strict mode (issue #299)."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/RenderingControl_1.xml")] = HttpResponse(
            200,
            {},
            read_file("dlna/dmr/RenderingControl_1_invalid_datetime_case.xml"),
        )

        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)

        with pytest.raises(UpnpError, match="Unsupported data type"):
            await factory.async_create_device("http://dlna_dmr:1234/device.xml")


class TestUpnpAction:
    """Tests for UpnpAction."""

    @pytest.mark.asyncio
    async def test_init(self) -> None:
        """Test Initializing a UpnpAction."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        assert action
        assert action.name == "GetVolume"

    @pytest.mark.asyncio
    async def test_valid_arguments(self) -> None:
        """Test validating arguments of an action."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("SetVolume")

        # all ok
        action.validate_arguments(InstanceID=0, Channel="Master", DesiredVolume=10)

        # invalid type for InstanceID
        try:
            action.validate_arguments(InstanceID="0", Channel="Master", DesiredVolume=10)
            assert False
        except UpnpValueError:
            pass

        # missing DesiredVolume
        try:
            action.validate_arguments(InstanceID="0", Channel="Master")
            assert False
        except UpnpValueError:
            pass

    @pytest.mark.asyncio
    async def test_format_request(self) -> None:
        """Test the request an action sends."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("SetVolume")

        service_type = "urn:schemas-upnp-org:service:RenderingControl:1"
        request = action.create_request(InstanceID=0, Channel="Master", DesiredVolume=10)

        root = DET.fromstring(request.body)
        namespace = {"rc_service": service_type}
        assert root.find(".//rc_service:SetVolume", namespace) is not None
        assert root.find(".//DesiredVolume", namespace) is not None

    @pytest.mark.asyncio
    async def test_format_request_escape(self) -> None:
        """Test escaping the request an action sends."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:AVTransport:1")
        action = service.action("SetAVTransportURI")

        service_type = "urn:schemas-upnp-org:service:AVTransport:1"
        metadata = "<item>test thing</item>"
        request = action.create_request(
            InstanceID=0,
            CurrentURI="http://example.org/file.mp3",
            CurrentURIMetaData=metadata,
        )

        root = DET.fromstring(request.body)
        namespace = {"avt_service": service_type}
        assert root.find(".//avt_service:SetAVTransportURI", namespace) is not None
        assert root.find(".//CurrentURIMetaData", namespace) is not None
        assert root.findtext(".//CurrentURIMetaData", None, namespace) == "<item>test thing</item>"

        current_uri_metadata_el = root.find(".//CurrentURIMetaData", namespace)
        assert current_uri_metadata_el is not None
        # This shouldn't have any children, due to its contents being escaped.
        assert current_uri_metadata_el.findall("./") == []

    @pytest.mark.asyncio
    async def test_parse_response(self) -> None:
        """Test calling an action and handling its response."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        service_type = "urn:schemas-upnp-org:service:RenderingControl:1"
        response_body = read_file("dlna/dmr/action_GetVolume.xml")
        response = HttpResponse(200, {}, response_body)
        result = action.parse_response(service_type, response)
        assert result == {"CurrentVolume": 3}

    @pytest.mark.asyncio
    async def test_parse_response_empty(self) -> None:
        """Test calling an action and handling an empty XML response."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("SetVolume")

        service_type = "urn:schemas-upnp-org:service:RenderingControl:1"
        response_body = read_file("dlna/dmr/action_SetVolume.xml")
        response = HttpResponse(200, {}, response_body)
        result = action.parse_response(service_type, response)
        assert result == {}

    @pytest.mark.asyncio
    async def test_parse_response_error(self) -> None:
        """Test calling and action and handling an invalid XML response."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        service_type = "urn:schemas-upnp-org:service:RenderingControl:1"
        response_body = read_file("dlna/dmr/action_GetVolumeError.xml")
        response = HttpResponse(200, {}, response_body)
        with pytest.raises(UpnpActionError) as exc:
            action.parse_response(service_type, response)
        assert exc.value.error_code == UpnpActionErrorCode.INVALID_ARGS
        assert exc.value.error_desc == "Invalid Args"

    @pytest.mark.asyncio
    async def test_parse_response_escape(self) -> None:
        """Test calling an action and properly (not) escaping the response."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:AVTransport:1")
        action = service.action("GetMediaInfo")

        service_type = "urn:schemas-upnp-org:service:AVTransport:1"
        response_body = read_file("dlna/dmr/action_GetMediaInfo.xml")
        response = HttpResponse(200, {}, response_body)
        result = action.parse_response(service_type, response)
        assert result == {
            "CurrentURI": "uri://1.mp3",
            "CurrentURIMetaData": "<DIDL-Lite "
            'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" '
            'xmlns:dc="http://purl.org/dc/elements/1.1/" '
            'xmlns:dlna="urn:schemas-dlna-org:metadata-1-0/" '
            'xmlns:sec="http://www.sec.co.kr/" '
            'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
            'xmlns:xbmc="urn:schemas-xbmc-org:metadata-1-0/">'
            '<item id="" parentID="" refID="" restricted="1">'
            "<upnp:artist>A &amp; B &gt; C</upnp:artist>"
            "</item>"
            "</DIDL-Lite>",
            "MediaDuration": "00:00:01",
            "NextURI": "",
            "NextURIMetaData": "",
            "NrTracks": 1,
            "PlayMedium": "NONE",
            "RecordMedium": "NOT_IMPLEMENTED",
            "WriteStatus": "NOT_IMPLEMENTED",
        }

    @pytest.mark.asyncio
    async def test_parse_response_no_service_type_version(self) -> None:
        """Test calling and action and handling a response without service type number."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        service_type = "urn:schemas-upnp-org:service:RenderingControl:1"
        response_body = read_file("dlna/dmr/action_GetVolumeInvalidServiceType.xml")
        response = HttpResponse(200, {}, response_body)
        try:
            action.parse_response(service_type, response)
            assert False
        except UpnpError:
            pass

    @pytest.mark.asyncio
    async def test_parse_response_no_service_type_version_2(self) -> None:
        """Test calling and action and handling a response without service type number."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:AVTransport:1")
        action = service.action("GetTransportInfo")

        service_type = "urn:schemas-upnp-org:service:AVTransport:1"
        response_body = read_file("dlna/dmr/action_GetTransportInfoInvalidServiceType.xml")
        response = HttpResponse(200, {}, response_body)
        try:
            action.parse_response(service_type, response)
            assert False
        except UpnpError:
            pass

    @pytest.mark.asyncio
    async def test_unknown_out_argument(self) -> None:
        """Test calling an action and handling an unknown out-argument."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        device_url = "http://dlna_dmr:1234/device.xml"
        service_type = "urn:schemas-upnp-org:service:RenderingControl:1"
        test_action = "GetVolume"

        factory = UpnpFactory(requester)
        device = await factory.async_create_device(device_url)
        service = device.service(service_type)
        action = service.action(test_action)

        response_body = read_file("dlna/dmr/action_GetVolumeExtraOutParameter.xml")
        response = HttpResponse(200, {}, response_body)
        try:
            action.parse_response(service_type, response)
            assert False
        except UpnpError:
            pass

        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device(device_url)
        service = device.service(service_type)
        action = service.action(test_action)

        try:
            action.parse_response(service_type, response)
        except UpnpError:
            assert False

    @pytest.mark.asyncio
    async def test_response_invalid_xml_namespaces(self) -> None:
        """Test parsing response with invalid XML namespaces."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        device_url = "http://igd:1234/device.xml"
        service_type = "urn:schemas-upnp-org:service:WANIPConnection:1"
        test_action = "DeletePortMapping"

        # Test strict mode.
        factory = UpnpFactory(requester)
        device = await factory.async_create_device(device_url)
        service = device.find_service(service_type)
        assert service is not None
        action = service.action(test_action)

        response_body = read_file("igd/action_WANPIPConnection_DeletePortMapping.xml")
        response = HttpResponse(200, {}, response_body)
        try:
            action.parse_response(service_type, response)
            assert False
        except UpnpError:
            pass

        # Test non-strict mode.
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device(device_url)
        service = device.find_service(service_type)
        assert service is not None
        action = service.action(test_action)

        try:
            action.parse_response(service_type, response)
        except UpnpError:
            assert False


class TestUpnpService:
    """Tests for UpnpService."""

    @pytest.mark.asyncio
    async def test_init(self) -> None:
        """Test initializing a UpnpService."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")

        base_url = "http://dlna_dmr:1234"
        assert service
        assert service.service_type == "urn:schemas-upnp-org:service:RenderingControl:1"
        assert service.control_url == base_url + "/upnp/control/RenderingControl1"
        assert service.event_sub_url == base_url + "/upnp/event/RenderingControl1"
        assert service.scpd_url == base_url + "/RenderingControl_1.xml"

    @pytest.mark.asyncio
    async def test_state_variables_actions(self) -> None:
        """Test eding a UpnpStateVariable."""
        requester = UpnpTestRequester(RESPONSE_MAP)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")

        state_var = service.state_variable("Volume")
        assert state_var

        action = service.action("GetVolume")
        assert action

    @pytest.mark.asyncio
    async def test_call_action(self) -> None:
        """Test calling a UpnpAction."""
        responses: MutableMapping = {
            (
                "POST",
                "http://dlna_dmr:1234/upnp/control/RenderingControl1",
            ): HttpResponse(
                200,
                {},
                read_file("dlna/dmr/action_GetVolume.xml"),
            )
        }
        responses.update(RESPONSE_MAP)
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        result = await service.async_call_action(action, InstanceID=0, Channel="Master")
        assert result["CurrentVolume"] == 3

    @pytest.mark.asyncio
    async def test_soap_fault_http_error(self) -> None:
        """Test an action response with HTTP error and SOAP fault raises exception."""
        responses: MutableMapping = {
            (
                "POST",
                "http://dlna_dmr:1234/upnp/control/RenderingControl1",
            ): HttpResponse(
                500,
                {},
                read_file("dlna/dmr/action_GetVolumeError.xml"),
            )
        }
        responses.update(RESPONSE_MAP)
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        with pytest.raises(UpnpActionResponseError) as exc:
            await service.async_call_action(action, InstanceID=0, Channel="Master")
        assert exc.value.error_code == UpnpActionErrorCode.INVALID_ARGS
        assert exc.value.error_desc == "Invalid Args"
        assert exc.value.status == 500

    @pytest.mark.asyncio
    async def test_http_error(self) -> None:
        """Test an action response with HTTP error and blank body raises exception."""
        responses: MutableMapping = {
            (
                "POST",
                "http://dlna_dmr:1234/upnp/control/RenderingControl1",
            ): HttpResponse(
                500,
                {},
                "",
            )
        }
        responses.update(RESPONSE_MAP)
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        with pytest.raises(UpnpResponseError) as exc:
            await service.async_call_action(action, InstanceID=0, Channel="Master")
        assert exc.value.status == 500

    @pytest.mark.asyncio
    async def test_soap_fault_http_ok(self) -> None:
        """Test an action response with HTTP OK but SOAP fault raises exception."""
        responses: MutableMapping = {
            (
                "POST",
                "http://dlna_dmr:1234/upnp/control/RenderingControl1",
            ): HttpResponse(
                200,
                {},
                read_file("dlna/dmr/action_GetVolumeError.xml"),
            )
        }
        responses.update(RESPONSE_MAP)
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        service = device.service("urn:schemas-upnp-org:service:RenderingControl:1")
        action = service.action("GetVolume")

        with pytest.raises(UpnpActionError) as exc:
            await service.async_call_action(action, InstanceID=0, Channel="Master")
        assert exc.value.error_code == UpnpActionErrorCode.INVALID_ARGS
        assert exc.value.error_desc == "Invalid Args"

    @pytest.mark.parametrize(
        "rc_doc",
        [
            "dlna/dmr/RenderingControl_1_bad_namespace.xml",  # Bad namespace
            "dlna/dmr/RenderingControl_1_bad_root_tag.xml",  # Wrong root tag
            "dlna/dmr/RenderingControl_1_missing_state_table.xml",  # Missing state table
        ],
    )
    @pytest.mark.asyncio
    async def test_bad_scpd_strict(self, rc_doc: str) -> None:
        """Test handling of bad service descriptions in strict mode."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/RenderingControl_1.xml")] = HttpResponse(
            200,
            {},
            read_file(rc_doc),
        )
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester)
        with pytest.raises(UpnpXmlContentError):
            await factory.async_create_device("http://dlna_dmr:1234/device.xml")

    @pytest.mark.parametrize(
        "rc_doc",
        [
            "dlna/dmr/RenderingControl_1_bad_namespace.xml",  # Bad namespace
            "dlna/dmr/RenderingControl_1_bad_root_tag.xml",  # Wrong root tag
            "dlna/dmr/RenderingControl_1_missing_state_table.xml",  # Missing state table
        ],
    )
    @pytest.mark.asyncio
    async def test_bad_scpd_non_strict_fails(self, rc_doc: str) -> None:
        """Test bad SCPD in non-strict mode."""
        responses = dict(RESPONSE_MAP)
        responses[("GET", "http://dlna_dmr:1234/RenderingControl_1.xml")] = HttpResponse(
            200,
            {},
            read_file(rc_doc),
        )
        requester = UpnpTestRequester(responses)
        factory = UpnpFactory(requester, non_strict=True)
        device = await factory.async_create_device("http://dlna_dmr:1234/device.xml")
        # Known good service
        assert device.services["urn:schemas-upnp-org:service:AVTransport:1"]
        # Bad service will also exist, to some extent
        assert device.services["urn:schemas-upnp-org:service:RenderingControl:1"]


class TestServiceUrlSsrf:
    """Tests that absolute service URLs to a foreign host are rejected (SSRF).

    Per UDA 1.1/2.0, SCPDURL/controlURL/eventSubURL must be relative to the
    device description URL. An absolute URL to a different host is treated as a
    potential SSRF vector.
    """

    DEVICE_URL = "http://192.168.1.50:1234/device.xml"
    SCPD_XML = (
        '<?xml version="1.0"?>'
        '<scpd xmlns="urn:schemas-upnp-org:service-1-0">'
        "<specVersion><major>1</major><minor>0</minor></specVersion>"
        "<actionList/><serviceStateTable/></scpd>"
    )

    def _device_xml(
        self,
        *,
        control_url: str = "/svc/control",
        scpd_url: str = "/svc.xml",
        event_sub_url: str = "/svc/event",
    ) -> str:
        """Build a single-service device description."""
        return (
            '<?xml version="1.0"?>'
            '<root xmlns="urn:schemas-upnp-org:device-1-0">'
            "<specVersion><major>1</major><minor>0</minor></specVersion>"
            "<device>"
            "<deviceType>urn:schemas-upnp-org:device:Foo:1</deviceType>"
            "<friendlyName>Foo</friendlyName>"
            "<UDN>uuid:00000000-0000-0000-0000-0000000000aa</UDN>"
            "<serviceList><service>"
            "<serviceType>urn:schemas-upnp-org:service:Bar:1</serviceType>"
            "<serviceId>urn:upnp-org:serviceId:Bar</serviceId>"
            f"<SCPDURL>{scpd_url}</SCPDURL>"
            f"<controlURL>{control_url}</controlURL>"
            f"<eventSubURL>{event_sub_url}</eventSubURL>"
            "</service></serviceList>"
            "</device></root>"
        )

    @pytest.mark.asyncio
    async def test_foreign_host_control_url_raises_in_strict_mode(self) -> None:
        """A controlURL pointing at another host raises in strict mode."""
        xml = self._device_xml(control_url="http://169.254.169.254/latest/meta-data/")
        requester = UpnpTestRequester({("GET", self.DEVICE_URL): HttpResponse(200, {}, xml)})
        factory = UpnpFactory(requester)

        with pytest.raises(UpnpError):
            await factory.async_create_device(self.DEVICE_URL)

    @pytest.mark.asyncio
    async def test_foreign_host_control_url_skipped_in_non_strict_mode(self) -> None:
        """A controlURL pointing at another host is skipped (not fetched) in non-strict mode."""
        xml = self._device_xml(control_url="http://169.254.169.254/latest/meta-data/")
        # No response registered for the foreign host: if it were requested, it would KeyError.
        requester = UpnpTestRequester({("GET", self.DEVICE_URL): HttpResponse(200, {}, xml)})
        factory = UpnpFactory(requester, non_strict=True)

        device = await factory.async_create_device(self.DEVICE_URL)

        assert device.services == {}

    @pytest.mark.asyncio
    async def test_malformed_control_url_raises_upnp_error_not_value_error(self) -> None:
        """A malformed service URL fails closed as UpnpError, not an unhandled ValueError."""
        xml = self._device_xml(control_url="http://[::1")  # unbalanced IPv6 bracket
        requester = UpnpTestRequester({("GET", self.DEVICE_URL): HttpResponse(200, {}, xml)})
        factory = UpnpFactory(requester)

        with pytest.raises(UpnpError):
            await factory.async_create_device(self.DEVICE_URL)

    @pytest.mark.asyncio
    async def test_malformed_control_url_skipped_in_non_strict_mode(self) -> None:
        """A malformed service URL is skipped (not raised) in non-strict mode."""
        xml = self._device_xml(control_url="http://[::1")  # unbalanced IPv6 bracket
        requester = UpnpTestRequester({("GET", self.DEVICE_URL): HttpResponse(200, {}, xml)})
        factory = UpnpFactory(requester, non_strict=True)

        device = await factory.async_create_device(self.DEVICE_URL)

        assert device.services == {}

    @pytest.mark.asyncio
    async def test_foreign_host_scpd_url_raises_in_strict_mode(self) -> None:
        """An SCPDURL pointing at another host raises before being fetched."""
        xml = self._device_xml(scpd_url="http://127.0.0.1:8123/api/states")
        requester = UpnpTestRequester({("GET", self.DEVICE_URL): HttpResponse(200, {}, xml)})
        factory = UpnpFactory(requester)

        with pytest.raises(UpnpError):
            await factory.async_create_device(self.DEVICE_URL)

    @pytest.mark.asyncio
    async def test_absolute_same_host_urls_allowed(self) -> None:
        """Absolute URLs to the device's own host are allowed (spec-compliant resolution)."""
        xml = self._device_xml(
            scpd_url="http://192.168.1.50:1234/svc.xml",
            control_url="http://192.168.1.50:1234/svc/control",
            event_sub_url="http://192.168.1.50:1234/svc/event",
        )
        requester = UpnpTestRequester(
            {
                ("GET", self.DEVICE_URL): HttpResponse(200, {}, xml),
                ("GET", "http://192.168.1.50:1234/svc.xml"): HttpResponse(200, {}, self.SCPD_XML),
            }
        )
        factory = UpnpFactory(requester, non_strict=True)

        device = await factory.async_create_device(self.DEVICE_URL)

        service = device.service("urn:schemas-upnp-org:service:Bar:1")
        assert service.control_url == "http://192.168.1.50:1234/svc/control"
        assert service.event_sub_url == "http://192.168.1.50:1234/svc/event"

    @pytest.mark.asyncio
    async def test_relative_urls_allowed(self) -> None:
        """Ordinary relative service URLs resolve against the device URL and are allowed."""
        xml = self._device_xml()
        requester = UpnpTestRequester(
            {
                ("GET", self.DEVICE_URL): HttpResponse(200, {}, xml),
                ("GET", "http://192.168.1.50:1234/svc.xml"): HttpResponse(200, {}, self.SCPD_XML),
            }
        )
        factory = UpnpFactory(requester, non_strict=True)

        device = await factory.async_create_device(self.DEVICE_URL)

        service = device.service("urn:schemas-upnp-org:service:Bar:1")
        assert service.control_url == "http://192.168.1.50:1234/svc/control"

    @pytest.mark.asyncio
    async def test_absolute_same_host_different_port_allowed(self) -> None:
        """An absolute URL to the device's host on a different port is allowed.

        The SSRF boundary is the host: a different port on the device's own host
        is the device talking to itself, not a cross-host request. Some devices
        legitimately serve control/eventing on a different port than the device
        description, so the host-only check intentionally permits this.
        """
        xml = self._device_xml(
            scpd_url="http://192.168.1.50:1234/svc.xml",
            control_url="http://192.168.1.50:8080/svc/control",
            event_sub_url="http://192.168.1.50:8080/svc/event",
        )
        requester = UpnpTestRequester(
            {
                ("GET", self.DEVICE_URL): HttpResponse(200, {}, xml),
                ("GET", "http://192.168.1.50:1234/svc.xml"): HttpResponse(200, {}, self.SCPD_XML),
            }
        )
        factory = UpnpFactory(requester, non_strict=True)

        device = await factory.async_create_device(self.DEVICE_URL)

        service = device.service("urn:schemas-upnp-org:service:Bar:1")
        assert service.control_url == "http://192.168.1.50:8080/svc/control"



"""async_upnp_client.aiohttp module."""

import asyncio
import logging
from asyncio.events import AbstractEventLoop, AbstractServer
from ipaddress import ip_address
from typing import Mapping
from urllib.parse import urlparse

import aiohttp.web
from aiohttp import (
    ClientConnectionError,
    ClientError,
    ClientResponseError,
    ClientSession,
    ClientTimeout,
)

from async_upnp_client.client import UpnpRequester
from async_upnp_client.const import (
    AddressTupleVXType,
    HttpRequest,
    HttpResponse,
    IPvXAddress,
)
from async_upnp_client.event_handler import UpnpEventHandler, UpnpNotifyServer
from async_upnp_client.exceptions import (
    UpnpClientResponseError,
    UpnpCommunicationError,
    UpnpConnectionError,
    UpnpConnectionTimeoutError,
    UpnpServerOSError,
)

_LOGGER = logging.getLogger(__name__)
_LOGGER_TRAFFIC_UPNP = logging.getLogger("async_upnp_client.traffic.upnp")


def _fixed_host_header(url: str) -> dict[str, str]:
    """Strip scope_id from IPv6 host, if needed."""
    if "%" not in url:
        return {}

    url_parts = urlparse(url)
    if url_parts.hostname and "%" in url_parts.hostname:
        idx = url_parts.hostname.rindex("%")
        fixed_hostname = url_parts.hostname[:idx]
        if ":" in fixed_hostname:
            fixed_hostname = f"[{fixed_hostname}]"
        host = f"{fixed_hostname}:{url_parts.port}" if url_parts.port else fixed_hostname
        return {"Host": host}

    return {}


class AiohttpRequester(UpnpRequester):
    """Standard AioHttpUpnpRequester, to be used with UpnpFactory."""

    # pylint: disable=too-few-public-methods

    def __init__(self, timeout: int = 5, http_headers: Mapping[str, str] | None = None) -> None:
        """Initialize."""
        self._timeout = ClientTimeout(total=float(timeout))
        self._http_headers = http_headers or {}

    async def async_http_request(
        self,
        http_request: HttpRequest,
    ) -> HttpResponse:
        """Do a HTTP request."""
        req_headers = {
            **_fixed_host_header(http_request.url),
            **self._http_headers,
            **(http_request.headers or {}),
        }

        log_traffic = _LOGGER_TRAFFIC_UPNP.isEnabledFor(logging.DEBUG)
        if log_traffic:  # pragma: no branch
            _LOGGER_TRAFFIC_UPNP.debug(
                "Sending request:\n%s %s\n%s\n%s\n",
                http_request.method,
                http_request.url,
                "\n".join([key + ": " + value for key, value in (req_headers or {}).items()]),
                http_request.body or "",
            )

        try:
            async with ClientSession() as session:
                async with session.request(
                    http_request.method,
                    http_request.url,
                    headers=req_headers,
                    data=http_request.body,
                    timeout=self._timeout,
                ) as response:
                    status = response.status
                    resp_headers: Mapping = response.headers or {}
                    resp_body = await response.read()

                    if log_traffic:  # pragma: no branch
                        _LOGGER_TRAFFIC_UPNP.debug(
                            "Got response from %s %s:\n%s\n%s\n\n%s",
                            http_request.method,
                            http_request.url,
                            status,
                            "\n".join([key + ": " + value for key, value in resp_headers.items()]),
                            resp_body,
                        )

                    resp_body_text = await response.text()
        except asyncio.TimeoutError as err:
            raise UpnpConnectionTimeoutError(repr(err)) from err
        except ClientConnectionError as err:
            raise UpnpConnectionError(repr(err)) from err
        except ClientResponseError as err:
            raise UpnpClientResponseError(
                request_info=err.request_info,
                history=err.history,
                status=err.status,
                message=err.message,
                headers=err.headers,
            ) from err
        except ClientError as err:
            raise UpnpCommunicationError(repr(err)) from err
        except UnicodeDecodeError as err:
            raise UpnpCommunicationError(repr(err)) from err

        return HttpResponse(status, resp_headers, resp_body_text)


class AiohttpSessionRequester(UpnpRequester):
    """
    Standard AiohttpSessionRequester, to be used with UpnpFactory.

    With pluggable session.
    """

    # pylint: disable=too-few-public-methods

    def __init__(
        self,
        session: ClientSession,
        with_sleep: bool = False,
        timeout: int = 5,
        http_headers: Mapping[str, str] | None = None,
    ) -> None:
        """Initialize."""
        self._session = session
        self._with_sleep = with_sleep
        self._timeout = ClientTimeout(total=float(timeout))
        self._http_headers = http_headers or {}

    async def async_http_request(
        self,
        http_request: HttpRequest,
    ) -> HttpResponse:
        """Do a HTTP request with a retry on ServerDisconnectedError.

        The HTTP/1.1 spec allows the server to disconnect at any time.
        We want to retry the request in this event.
        """
        for _ in range(2):
            try:
                return await self._async_http_request(http_request)
            except ClientConnectionError as err:
                _LOGGER.debug(
                    "%r during request %s %s; retrying",
                    err,
                    http_request.method,
                    http_request.url,
                )
        try:
            return await self._async_http_request(http_request)
        except ClientConnectionError as err:
            raise UpnpConnectionError(repr(err)) from err

    async def _async_http_request(
        self,
        http_request: HttpRequest,
    ) -> HttpResponse:
        """Do a HTTP request."""
        # pylint: disable=too-many-arguments
        req_headers = {
            **_fixed_host_header(http_request.url),
            **self._http_headers,
            **(http_request.headers or {}),
        }

        log_traffic = _LOGGER_TRAFFIC_UPNP.isEnabledFor(logging.DEBUG)
        if log_traffic:  # pragma: no branch
            _LOGGER_TRAFFIC_UPNP.debug(
                "Sending request:\n%s %s\n%s\n%s\n",
                http_request.method,
                http_request.url,
                "\n".join([key + ": " + value for key, value in (req_headers or {}).items()]),
                http_request.body or "",
            )

        if self._with_sleep:
            await asyncio.sleep(0)

        try:
            async with self._session.request(
                http_request.method,
                http_request.url,
                headers=req_headers,
                data=http_request.body,
                timeout=self._timeout,
            ) as response:
                status = response.status
                resp_headers: Mapping = response.headers or {}
                resp_body = await response.read()

                if log_traffic:  # pragma: no branch
                    _LOGGER_TRAFFIC_UPNP.debug(
                        "Got response from %s %s:\n%s\n%s\n\n%s",
                        http_request.method,
                        http_request.url,
                        status,
                        "\n".join([key + ": " + value for key, value in resp_headers.items()]),
                        resp_body,
                    )

                resp_body_text = await response.text()
        except asyncio.TimeoutError as err:
            raise UpnpConnectionTimeoutError(repr(err)) from err
        except ClientConnectionError:
            raise
        except ClientResponseError as err:
            raise UpnpClientResponseError(
                request_info=err.request_info,
                history=err.history,
                status=err.status,
                message=err.message,
                headers=err.headers,
            ) from err
        except ClientError as err:
            raise UpnpCommunicationError(repr(err)) from err
        except UnicodeDecodeError as err:
            raise UpnpCommunicationError(repr(err)) from err

        return HttpResponse(status, resp_headers, resp_body_text)


class AiohttpNotifyServer(UpnpNotifyServer):
    """
    Aio HTTP Server to handle incoming events.

    It is advisable to use one AiohttpNotifyServer per listening IP,
    UpnpDevices can share a AiohttpNotifyServer/UpnpEventHandler.
    """

    def __init__(
        self,
        requester: UpnpRequester,
        source: AddressTupleVXType,
        callback_url: str | None = None,
        loop: AbstractEventLoop | None = None,
    ) -> None:
        """Initialize."""
        self._source = source
        self._callback_url = callback_url
        self._loop = loop or asyncio.get_event_loop()

        self._aiohttp_server: aiohttp.web.Server | None = None
        self._server: AbstractServer | None = None

        self.event_handler = UpnpEventHandler(self, requester)

    async def async_start_server(self) -> None:
        """Start the HTTP server."""
        self._aiohttp_server = aiohttp.web.Server(self._handle_request)

        try:
            self._server = await self._loop.create_server(self._aiohttp_server, self._source[0], self._source[1])
        except OSError as err:
            _LOGGER.error(
                "Failed to create HTTP server at %s:%d: %s",
                self._source[0],
                self._source[1],
                err,
            )
            raise UpnpServerOSError(
                errno=err.errno,
                strerror=err.strerror,
            ) from err

        # Get listening port.
        socks = self._server.sockets
        assert socks and len(socks) == 1
        sock = socks[0]
        self._source = sock.getsockname()
        _LOGGER.debug("New source for UpnpNotifyServer: %s", self._source)

    async def async_stop_server(self) -> None:
        """Stop the HTTP server."""
        await self.event_handler.async_unsubscribe_all()

        if self._aiohttp_server:
            await self._aiohttp_server.shutdown(10)
            self._aiohttp_server = None

        if self._server:
            self._server.close()
            self._server = None

    async def _handle_request(self, request: aiohttp.web.BaseRequest) -> aiohttp.web.Response:
        """Handle incoming requests."""
        _LOGGER.debug("Received request: %s", request)
        log_traffic = _LOGGER_TRAFFIC_UPNP.isEnabledFor(logging.DEBUG)

        headers = request.headers
        body = await request.text()
        if log_traffic:
            _LOGGER_TRAFFIC_UPNP.debug(
                "Incoming request:\nNOTIFY\n%s\n\n%s",
                "\n".join([key + ": " + value for key, value in headers.items()]),
                body,
            )

        if request.method != "NOTIFY":
            _LOGGER.debug("Not notify")
            return aiohttp.web.Response(status=405)

        http_request = HttpRequest(request.method, self.callback_url, request.headers, body)
        status = await self.event_handler.handle_notify(http_request)
        _LOGGER.debug("NOTIFY response status: %s", status)
        if log_traffic:
            _LOGGER_TRAFFIC_UPNP.debug("Sending response: %s", status)

        return aiohttp.web.Response(status=status)

    @property
    def listen_ip(self) -> IPvXAddress:
        """Get listening IP Address."""
        return ip_address(self._source[0])

    @property
    def listen_host(self) -> str:
        """Get listening host."""
        return str(self.listen_ip)

    @property
    def listen_port(self) -> int:
        """Get the listening port."""
        return self._source[1]

    @property
    def callback_url(self) -> str:
        """Return callback URL on which we are callable."""
        listen_ip = self.listen_ip
        return self._callback_url or (
            self._callback_url or f"http://{self.listen_host}:{self.listen_port}/notify"
            if listen_ip.version == 4
            else f"http://[{self.listen_host}]:{self.listen_port}/notify"
        )



"""Unit tests for ssdp."""

import asyncio
from unittest.mock import ANY, AsyncMock, MagicMock

import pytest

from async_upnp_client.ssdp import (
    SSDP_PORT,
    SsdpProtocol,
    build_ssdp_search_packet,
    decode_ssdp_packet,
    fix_ipv6_address_scope_id,
    get_ssdp_socket,
    is_ipv4_address,
    is_ipv6_address,
    is_valid_ssdp_packet,
)


def test_ssdp_search_packet() -> None:
    """Test SSDP search packet generation."""
    msg = build_ssdp_search_packet(("239.255.255.250", 1900), 4, "ssdp:all")
    assert (
        msg
        == "M-SEARCH * HTTP/1.1\r\n"
        "HOST:239.255.255.250:1900\r\n"
        'MAN:"ssdp:discover"\r\n'
        "MX:4\r\n"
        "ST:ssdp:all\r\n"
        "\r\n".encode()
    )


def test_ssdp_search_packet_v6() -> None:
    """Test SSDP search packet generation."""
    msg = build_ssdp_search_packet(("FF02::C", 1900, 0, 2), 4, "ssdp:all")
    assert (
        msg
        == "M-SEARCH * HTTP/1.1\r\n"
        "HOST:[FF02::C%2]:1900\r\n"
        'MAN:"ssdp:discover"\r\n'
        "MX:4\r\n"
        "ST:ssdp:all\r\n"
        "\r\n".encode()
    )


def test_is_valid_ssdp_packet() -> None:
    """Test SSDP response validation."""
    assert not is_valid_ssdp_packet(b"")

    msg = (
        b"HTTP/1.1 200 OK\r\n"
        b"Cache-Control: max-age=1900\r\n"
        b"Location: http://192.168.1.1:80/RootDevice.xml\r\n"
        b"Server: UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0\r\n"
        b"ST:urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1\r\n"
        b"USN: uuid:...::WANCommonInterfaceConfig:1\r\n"
        b"EXT:\r\n"
        b"\r\n"
    )
    assert is_valid_ssdp_packet(msg)


def test_decode_ssdp_packet() -> None:
    """Test SSDP response decoding."""
    msg = (
        b"HTTP/1.1 200 OK\r\n"
        b"Cache-Control: max-age=1900\r\n"
        b"Location: http://192.168.1.1:80/RootDevice.xml\r\n"
        b"Server: UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0\r\n"
        b"ST:urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1\r\n"
        b"USN: uuid:...::WANCommonInterfaceConfig:1\r\n"
        b"EXT:\r\n"
        b"\r\n"
    )
    request_line, headers = decode_ssdp_packet(msg, ("local_addr", 1900), ("remote_addr", 12345))

    assert request_line == "HTTP/1.1 200 OK"

    assert headers == {
        "cache-control": "max-age=1900",
        "location": "http://192.168.1.1:80/RootDevice.xml",
        "server": "UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0",
        "st": "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        "usn": "uuid:...::WANCommonInterfaceConfig:1",
        "ext": "",
        "_location_original": "http://192.168.1.1:80/RootDevice.xml",
        "_host": "remote_addr",
        "_port": 12345,
        "_local_addr": ("local_addr", 1900),
        "_remote_addr": ("remote_addr", 12345),
        "_udn": "uuid:...",
        "_timestamp": ANY,
    }


def test_decode_ssdp_packet_missing_ending() -> None:
    """Test SSDP response decoding with a missing end line."""
    msg = (
        b"HTTP/1.1 200 OK\r\n"
        b"CACHE-CONTROL: max-age = 1800\r\n"
        b"DATE:Sun, 25 Apr 2021 16:08:06 GMT\r\n"
        b"EXT:\r\n"
        b"LOCATION: http://192.168.107.148:8088/description\r\n"
        b"SERVER: Ubuntu/10.04 UPnP/1.1 Harmony/16.3\r\n"
        b"ST: urn:myharmony-com:device:harmony:1\r\n"
        b"USN: uuid:...::urn:myharmony-com:device:harmony:1\r\n"
        b"BOOTID.UPNP.ORG:1619366886\r\n"
    )
    request_line, headers = decode_ssdp_packet(msg, ("local_addr", 1900), ("remote_addr", 12345))

    assert request_line == "HTTP/1.1 200 OK"

    assert headers == {
        "cache-control": "max-age = 1800",
        "date": "Sun, 25 Apr 2021 16:08:06 GMT",
        "location": "http://192.168.107.148:8088/description",
        "server": "Ubuntu/10.04 UPnP/1.1 Harmony/16.3",
        "st": "urn:myharmony-com:device:harmony:1",
        "usn": "uuid:...::urn:myharmony-com:device:harmony:1",
        "bootid.upnp.org": "1619366886",
        "ext": "",
        "_location_original": "http://192.168.107.148:8088/description",
        "_host": "remote_addr",
        "_port": 12345,
        "_local_addr": ("local_addr", 1900),
        "_remote_addr": ("remote_addr", 12345),
        "_udn": "uuid:...",
        "_timestamp": ANY,
    }


def test_decode_ssdp_packet_duplicate_header() -> None:
    """Test SSDP response decoding with a duplicate header."""
    msg = b"HTTP/1.1 200 OK\r\nCACHE-CONTROL: max-age = 1800\r\nCACHE-CONTROL: max-age = 1800\r\n\r\n"
    _, headers = decode_ssdp_packet(msg, ("local_addr", 1900), ("remote_addr", 12345))

    assert headers == {
        "cache-control": "max-age = 1800",
        "_host": "remote_addr",
        "_port": 12345,
        "_local_addr": ("local_addr", 1900),
        "_remote_addr": ("remote_addr", 12345),
        "_timestamp": ANY,
    }


def test_decode_ssdp_packet_empty_location() -> None:
    """Test SSDP response decoding with an empty location."""
    msg = b"HTTP/1.1 200 OK\r\nLOCATION: \r\nCACHE-CONTROL: max-age = 1800\r\n\r\n"
    _, headers = decode_ssdp_packet(msg, ("local_addr", 1900), ("remote_addr", 12345))

    assert headers == {
        "cache-control": "max-age = 1800",
        "location": "",
        "_host": "remote_addr",
        "_port": 12345,
        "_local_addr": ("local_addr", 1900),
        "_remote_addr": ("remote_addr", 12345),
        "_timestamp": ANY,
    }


@pytest.mark.asyncio
async def test_ssdp_protocol_handles_broken_headers() -> None:
    """Test SsdpProtocol is able to handle broken headers."""
    msg = b"HTTP/1.1 200 OK\r\nDEFUNCT\r\nCACHE-CONTROL: max-age = 1800\r\n\r\n"
    addr = ("addr", 123)
    loop = asyncio.get_event_loop()

    async_on_data_mock = AsyncMock()
    protocol = SsdpProtocol(loop, async_on_data=async_on_data_mock)
    protocol.transport = MagicMock()
    protocol.datagram_received(msg, addr)
    async_on_data_mock.assert_not_awaited()


def test_decode_ssdp_packet_v6() -> None:
    """Test SSDP response decoding."""
    msg = (
        b"HTTP/1.1 200 OK\r\n"
        b"Cache-Control: max-age=1900\r\n"
        b"Location: http://[fe80::2]:80/RootDevice.xml\r\n"
        b"Server: UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0\r\n"
        b"ST:urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1\r\n"
        b"USN: uuid:...::WANCommonInterfaceConfig:1\r\n"
        b"EXT:\r\n"
        b"\r\n"
    )

    request_line, headers = decode_ssdp_packet(msg, ("FF02::C", 1900, 0, 3), ("fe80::1", 123, 0, 3))

    assert request_line == "HTTP/1.1 200 OK"

    assert headers == {
        "cache-control": "max-age=1900",
        "location": "http://[fe80::2%3]:80/RootDevice.xml",
        "server": "UPnP/1.0 UPnP/1.0 UPnP-Device-Host/1.0",
        "st": "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
        "usn": "uuid:...::WANCommonInterfaceConfig:1",
        "ext": "",
        "_location_original": "http://[fe80::2]:80/RootDevice.xml",
        "_host": "fe80::1%3",
        "_port": 123,
        "_local_addr": ("FF02::C", 1900, 0, 3),
        "_remote_addr": ("fe80::1", 123, 0, 3),
        "_udn": "uuid:...",
        "_timestamp": ANY,
    }


def test_get_ssdp_socket() -> None:
    """Test get_ssdp_socket accepts a port."""
    # Without a port, should default to SSDP_PORT
    _, source, target = get_ssdp_socket(("127.0.0.1", 0), ("127.0.0.1", SSDP_PORT))
    assert source == ("127.0.0.1", 0)
    assert target == ("127.0.0.1", SSDP_PORT)

    # With a different port.
    _, source, target = get_ssdp_socket(
        ("127.0.0.1", 0),
        ("127.0.0.1", 1234),
    )
    assert source == ("127.0.0.1", 0)
    assert target == ("127.0.0.1", 1234)


def test_microsoft_butchers_ssdp() -> None:
    """Test parsing a `Microsoft Windows Peer Name Resolution Protocol` packet."""
    msg = (
        b"HTTP/1.1 200 OK\r\n"
        b"ST:urn:Microsoft Windows Peer Name Resolution Protocol: V4:IPV6:LinkLocal\r\n"
        b"USN:[fe80::aaaa:bbbb:cccc:dddd]:3540\r\n"
        b"Location:192.168.1.1\r\n"
        b"AL:[fe80::aaaa:bbbb:cccc:dddd]:3540\r\n"
        b'OPT:"http://schemas.upnp.org/upnp/1/0/"; ns=01\r\n'
        b"01-NLS:abcdef0123456789abcdef012345678\r\n"
        b"Cache-Control:max-age=14400\r\n"
        b"Server:Microsoft-Windows-NT/5.1 UPnP/1.0 UPnP-Device-Host/1.0\r\n"
        b"Ext:\r\n"
    )

    request_line, headers = decode_ssdp_packet(msg, ("239.255.255.250", 1900), ("192.168.1.1", 12345))

    assert request_line == "HTTP/1.1 200 OK"
    assert headers == {
        "st": "urn:Microsoft Windows Peer Name Resolution Protocol: V4:IPV6:LinkLocal",
        "usn": "[fe80::aaaa:bbbb:cccc:dddd]:3540",
        "location": "192.168.1.1",
        "al": "[fe80::aaaa:bbbb:cccc:dddd]:3540",
        "opt": '"http://schemas.upnp.org/upnp/1/0/"; ns=01',
        "01-nls": "abcdef0123456789abcdef012345678",
        "cache-control": "max-age=14400",
        "server": "Microsoft-Windows-NT/5.1 UPnP/1.0 UPnP-Device-Host/1.0",
        "ext": "",
        "_location_original": "192.168.1.1",
        "_host": "192.168.1.1",
        "_port": 12345,
        "_local_addr": ("239.255.255.250", 1900),
        "_remote_addr": ("192.168.1.1", 12345),
        "_timestamp": ANY,
    }


def test_is_ipv4_address() -> None:
    """Test is_ipv4_address()."""
    assert is_ipv4_address(("192.168.1.1", 12345))
    assert not is_ipv4_address(("fe80::1", 12345, 0, 6))


def test_is_ipv6_address() -> None:
    """Test is_ipv6_address()."""
    assert is_ipv6_address(("fe80::1", 12345, 0, 6))
    assert not is_ipv6_address(("192.168.1.1", 12345))


def test_fix_ipv6_address_scope_id() -> None:
    """Test fix_ipv6_address_scope_id."""
    assert fix_ipv6_address_scope_id(("fe80::1", 0, 0, 4)) == ("fe80::1", 0, 0, 4)
    assert fix_ipv6_address_scope_id(("fe80::1%4", 0, 0, 4)) == ("fe80::1", 0, 0, 4)
    assert fix_ipv6_address_scope_id(("fe80::1%4", 0, 0, 0)) == ("fe80::1", 0, 0, 4)
    assert fix_ipv6_address_scope_id(None) is None
    assert fix_ipv6_address_scope_id(("192.168.1.1", 0)) == ("192.168.1.1", 0)



"""async_upnp_client.exceptions module."""

import asyncio
from enum import IntEnum
from typing import Any
from xml.etree import ElementTree as ET

import aiohttp

# pylint: disable=too-many-ancestors


class UpnpError(Exception):
    """Base class for all errors raised by this library."""

    def __init__(self, *args: Any, message: str | None = None, **_kwargs: Any) -> None:
        """Initialize base UpnpError."""
        super().__init__(*args, message)


class UpnpContentError(UpnpError):
    """Content of UPnP response is invalid."""


class UpnpActionErrorCode(IntEnum):
    """Error codes for UPnP Action errors."""

    INVALID_ACTION = 401
    INVALID_ARGS = 402
    # (DO_NOT_USE) = 403
    ACTION_FAILED = 501
    ARGUMENT_VALUE_INVALID = 600
    ARGUMENT_VALUE_OUT_OF_RANGE = 601
    OPTIONAL_ACTION_NOT_IMPLEMENTED = 602
    OUT_OF_MEMORY = 603
    HUMAN_INTERVENTION_REQUIRED = 604
    STRING_ARGUMENT_TOO_LONG = 605


class UpnpActionError(UpnpError):
    """Server returned a SOAP Fault in response to an Action."""

    def __init__(
        self,
        *args: Any,
        error_code: int | None = None,
        error_desc: str | None = None,
        message: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize from response body."""
        if not message:
            message = f"Received UPnP error {error_code} ({error_desc})"
        super().__init__(*args, message=message, **kwargs)
        self.error_code = error_code
        self.error_desc = error_desc


class UpnpXmlParseError(UpnpContentError, ET.ParseError):
    """UPnP response is not valid XML."""

    def __init__(self, orig_err: ET.ParseError) -> None:
        """Initialize from original ParseError, to match it."""
        super().__init__(message=str(orig_err))
        self.code = orig_err.code
        self.position = orig_err.position


class UpnpValueError(UpnpContentError):
    """Invalid value error."""

    def __init__(self, name: str, value: Any) -> None:
        """Initialize."""
        super().__init__(message=f"Invalid value for {name}: '{value}'")
        self.name = name
        self.value = value


class UpnpSIDError(UpnpContentError):
    """Missing Subscription Identifier from response."""


class UpnpXmlContentError(UpnpContentError):
    """XML document does not have expected content."""


class UpnpCommunicationError(UpnpError, aiohttp.ClientError):
    """Error occurred while communicating with the UPnP device ."""


class UpnpResponseError(UpnpCommunicationError):
    """HTTP error code returned by the UPnP device."""

    def __init__(
        self,
        *args: Any,
        status: int,
        headers: aiohttp.typedefs.LooseHeaders | None = None,
        message: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize."""
        if not message:
            message = f"Did not receive HTTP 200 but {status}"
        super().__init__(*args, message=message, **kwargs)
        self.status = status
        self.headers = headers


class UpnpActionResponseError(UpnpActionError, UpnpResponseError):
    """HTTP error code and UPnP error code.

    UPnP errors are usually indicated with HTTP 500 (Internal Server Error) and
    actual details in the response body as a SOAP Fault.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        *args: Any,
        status: int,
        headers: aiohttp.typedefs.LooseHeaders | None = None,
        error_code: int | None = None,
        error_desc: str | None = None,
        message: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize."""
        if not message:
            message = f"Received HTTP error code {status}, UPnP error code {error_code} ({error_desc})"
        super().__init__(
            *args,
            status=status,
            headers=headers,
            error_code=error_code,
            error_desc=error_desc,
            message=message,
            **kwargs,
        )


class UpnpClientResponseError(aiohttp.ClientResponseError, UpnpResponseError):  # type: ignore
    """HTTP response error with more details from aiohttp."""


class UpnpConnectionError(UpnpCommunicationError, aiohttp.ClientConnectionError):
    """Error in the underlying connection to the UPnP device.

    This could indicate that the device is offline.
    """


class UpnpConnectionTimeoutError(UpnpConnectionError, aiohttp.ServerTimeoutError, asyncio.TimeoutError):
    """Timeout while communicating with the device."""


class UpnpServerError(UpnpError):
    """Error with a local server."""


class UpnpServerOSError(UpnpServerError, OSError):
    """System-related error when starting a local server."""

    def __init___(self, errno: int, strerror: str) -> None:
        """Initialize simplified version of OSError."""
        OSError.__init__(self, errno, strerror)



"""async_upnp_client.ssdp_listener module."""

import asyncio
import logging
import re
import socket
from asyncio.events import AbstractEventLoop
from contextlib import suppress
from datetime import datetime, timedelta
from functools import lru_cache
from ipaddress import IPv4Address, IPv6Address, ip_address
from typing import TYPE_CHECKING, Any, Callable, Coroutine, KeysView, Mapping
from urllib.parse import unquote, urlparse

from async_upnp_client.advertisement import SsdpAdvertisementListener
from async_upnp_client.const import (
    AddressTupleVXType,
    DeviceOrServiceType,
    NotificationSubType,
    NotificationType,
    SearchTarget,
    SsdpSource,
    UniqueDeviceName,
)
from async_upnp_client.search import SsdpSearchListener
from async_upnp_client.ssdp import (
    SSDP_MX,
    SSDP_ST_ALL,
    determine_source_target,
    udn_from_usn,
)
from async_upnp_client.utils import CaseInsensitiveDict

_SENTINEL = object()
_LOGGER = logging.getLogger(__name__)
CACHE_CONTROL_RE = re.compile(r"max-age\s*=\s*(\d+)", re.IGNORECASE)
DEFAULT_MAX_AGE = timedelta(seconds=900)
IGNORED_HEADERS = {
    "date",
    "cache-control",
    "server",
    "host",
    "location",  # Location-header is handled differently!
}


# Hostnames that are reserved or conventionally resolve to loopback. RFC 6761
# additionally reserves the whole ``.localhost`` TLD, handled separately.
_LOOPBACK_HOSTNAMES = frozenset({"localhost", "localhost.localdomain"})


def _parse_ip_literal(hostname: str) -> IPv4Address | IPv6Address | None:
    """Parse hostname as an IP address, normalising obfuscated IPv4 encodings.

    Returns the address, or None if hostname is a DNS name. Besides the usual
    dotted-quad and standard IPv6 literals, this recognises the alternative
    numeric IPv4 encodings the OS resolver accepts -- decimal, octal, hex and
    short forms such as ``2130706433``, ``0177.0.0.1`` or ``127.1`` -- as well
    as percent-encoded variants, so that a loopback or link-local address cannot
    be smuggled past the filter in an obfuscated form. Real LAN devices never
    advertise these encodings, so normalising them carries no false-positive
    risk.
    """
    hostname = unquote(hostname)
    try:
        return ip_address(hostname)
    except ValueError:
        pass

    # ip_address() only accepts dotted-quad IPv4; fall back to inet_aton, which
    # interprets decimal/octal/hex/short forms the same way the resolver will.
    try:
        packed = socket.inet_aton(hostname)
    except OSError:
        # Not an IP literal in any encoding, but a DNS name.
        return None
    return IPv4Address(packed)


@lru_cache(maxsize=128)
def is_valid_location(location: str) -> bool:
    """Validate if this location is usable.

    Rejects loopback, unspecified and IPv4 link-local (169.254.0.0/16, which
    includes cloud-metadata endpoints) addresses. Uses proper URL/IP parsing
    rather than substring matching, so numeric, obfuscated (decimal/octal/hex)
    and IPv4-mapped-IPv6 forms of those addresses are caught too. A trailing
    root-label dot is normalised away, and the reserved loopback names
    ``localhost``, any ``*.localhost`` name (RFC 6761) and ``localhost.localdomain``
    are rejected by name. RFC1918/ULA and IPv6 link-local (fe80::/10) addresses
    are allowed, as those are where regular LAN UPnP devices live.

    This is best-effort defence-in-depth, not a security boundary: an arbitrary
    DNS name that resolves to a loopback/link-local address (including via DNS
    rebinding) is allowed here and can only be caught at connect time.
    """
    if not location.startswith("http"):
        return False

    try:
        hostname = urlparse(location).hostname
    except ValueError:
        # Malformed URL (e.g. an unbalanced IPv6 bracket); reject it.
        return False
    # Normalise a trailing root-label dot, e.g. "127.0.0.1." or "localhost.",
    # then reject reserved/conventional loopback names (no DNS resolution needed).
    hostname = (hostname or "").rstrip(".")
    if not hostname or hostname in _LOOPBACK_HOSTNAMES or hostname.endswith(".localhost"):
        return False

    ip_addr = _parse_ip_literal(hostname)
    if ip_addr is None:
        # Not an IP literal, but a DNS name; allow it.
        return True

    # Unwrap IPv4-mapped IPv6 (e.g. ::ffff:127.0.0.1) to its embedded IPv4.
    ip_addr = getattr(ip_addr, "ipv4_mapped", None) or ip_addr
    if ip_addr.is_loopback or ip_addr.is_unspecified:
        return False

    # Only IPv4 link-local is rejected (cloud metadata); IPv6 link-local
    # (fe80::/10) is legitimately used by LAN UPnP devices.
    return not (ip_addr.version == 4 and ip_addr.is_link_local)


def valid_search_headers(headers: CaseInsensitiveDict) -> bool:
    """Validate if this search is usable."""
    return headers.lower_values_true(("_udn", "st")) and is_valid_location(headers.get_lower("location", ""))


def valid_advertisement_headers(headers: CaseInsensitiveDict) -> bool:
    """Validate if this advertisement is usable for connecting to a device."""
    return headers.lower_values_true(("_udn", "nt", "nts")) and is_valid_location(headers.get_lower("location", ""))


def valid_byebye_headers(headers: CaseInsensitiveDict) -> bool:
    """Validate if this advertisement has required headers for byebye."""
    return headers.lower_values_true(("_udn", "nt", "nts"))


@lru_cache(maxsize=128)
def extract_uncache_after(cache_control: str) -> timedelta:
    """Get uncache after from cache control header."""
    if match := CACHE_CONTROL_RE.search(cache_control):
        max_age = int(match[1])
        return timedelta(seconds=max_age)
    return DEFAULT_MAX_AGE


def extract_valid_to(headers: CaseInsensitiveDict) -> datetime:
    """Extract/create valid to."""
    uncache_after = extract_uncache_after(headers.get_lower("cache-control", ""))
    timestamp: datetime = headers.get_lower("_timestamp")
    return timestamp + uncache_after


class SsdpDevice:
    """
    SSDP Device.

    Holds all known information about the device.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, udn: str, valid_to: datetime):
        """Initialize."""
        self.udn = udn
        self.valid_to: datetime = valid_to
        self._locations: dict[str, datetime] = {}
        self.last_seen: datetime | None = None
        self.search_headers: dict[DeviceOrServiceType, CaseInsensitiveDict] = {}
        self.advertisement_headers: dict[DeviceOrServiceType, CaseInsensitiveDict] = {}
        self.userdata: Any = None

    def add_location(self, location: str, valid_to: datetime) -> None:
        """Add a (new) location the device can be reached at."""
        self._locations[location] = valid_to

    @property
    def location(self) -> str | None:
        """
        Get a location of the device.

        Kept for compatibility, use method `locations`.
        """
        # Sort such that the same location will be given each time.
        for location in sorted(self.locations):
            return location

        return None

    @property
    def locations(self) -> KeysView[str]:
        """Get all know locations of the device."""
        return self._locations.keys()

    def purge_locations(self, now: datetime | None = None) -> None:
        """Purge locations which are no longer valid/timed out."""
        if not now:
            now = datetime.now()
        to_remove = [location for location, valid_to in self._locations.items() if now > valid_to]
        for location in to_remove:
            del self._locations[location]

    def combined_headers(
        self,
        device_or_service_type: DeviceOrServiceType,
    ) -> CaseInsensitiveDict:
        """Get headers from search and advertisement for a given device- or service type.

        If there are both search and advertisement headers,
        the search headers are combined with the advertisement headers and a new
        CaseInsensitiveDict is returned.

        If there are only search headers, the search headers are returned.

        If there are only advertisement headers, the advertisement headers are returned.

        If there are no headers, an empty CaseInsensitiveDict is returned.

        Callers should be aware that the returned CaseInsensitiveDict may be a view
        into the internal data structures of this class. If the caller modifies the
        returned CaseInsensitiveDict, the internal data structures will be modified
        as well.
        """
        search_headers = self.search_headers.get(device_or_service_type, _SENTINEL)
        advertisement_headers = self.advertisement_headers.get(device_or_service_type, _SENTINEL)
        if search_headers is not _SENTINEL and advertisement_headers is not _SENTINEL:
            if TYPE_CHECKING:
                assert isinstance(search_headers, CaseInsensitiveDict)
                assert isinstance(advertisement_headers, CaseInsensitiveDict)
            header_dict = search_headers.combine(advertisement_headers)
            header_dict.del_lower("_source")
            return header_dict
        if search_headers is not _SENTINEL:
            if TYPE_CHECKING:
                assert isinstance(search_headers, CaseInsensitiveDict)
            return search_headers
        if advertisement_headers is not _SENTINEL:
            if TYPE_CHECKING:
                assert isinstance(advertisement_headers, CaseInsensitiveDict)
            return advertisement_headers
        return CaseInsensitiveDict()

    @property
    def all_combined_headers(self) -> Mapping[DeviceOrServiceType, CaseInsensitiveDict]:
        """Get all headers from search and advertisement for all known device- and service types."""
        dsts = set(self.advertisement_headers).union(set(self.search_headers))
        return {dst: self.combined_headers(dst) for dst in dsts}

    def __repr__(self) -> str:
        """Return the representation."""
        return f"<{type(self).__name__}({self.udn})>"


def same_headers_differ(current_headers: CaseInsensitiveDict, new_headers: CaseInsensitiveDict) -> bool:
    """Compare headers present in both to see if anything interesting has changed."""
    current_headers_dict = current_headers.as_dict()
    new_headers_dict = new_headers.as_dict()
    new_headers_case_map = new_headers.case_map()
    current_headers_case_map = current_headers.case_map()

    for lower_header, current_header in current_headers_case_map.items():
        if (lower_header != "" and lower_header[0] == "_") or lower_header in IGNORED_HEADERS:
            continue
        new_header = new_headers_case_map.get(lower_header, _SENTINEL)

        if new_header is not _SENTINEL:
            current_value = current_headers_dict[current_header]
            new_value = new_headers_dict[new_header]  # type: ignore[index]
            if current_value != new_value:
                _LOGGER.debug(
                    "Header %s changed from %s to %s",
                    current_header,
                    current_value,
                    new_value,
                )
                return True

    return False


def headers_differ_from_existing_advertisement(
    ssdp_device: SsdpDevice, dst: DeviceOrServiceType, headers: CaseInsensitiveDict
) -> bool:
    """Compare against existing advertisement headers to see if anything interesting has changed."""
    headers_old = ssdp_device.advertisement_headers.get(dst, _SENTINEL)
    if headers_old is _SENTINEL:
        return False
    if TYPE_CHECKING:
        assert isinstance(headers_old, CaseInsensitiveDict)
    return same_headers_differ(headers_old, headers)


def headers_differ_from_existing_search(
    ssdp_device: SsdpDevice, dst: DeviceOrServiceType, headers: CaseInsensitiveDict
) -> bool:
    """Compare against existing search headers to see if anything interesting has changed."""
    headers_old = ssdp_device.search_headers.get(dst, _SENTINEL)
    if headers_old is _SENTINEL:
        return False
    if TYPE_CHECKING:
        assert isinstance(headers_old, CaseInsensitiveDict)
    return same_headers_differ(headers_old, headers)


def ip_version_from_location(location: str) -> int | None:
    """Get the ip version for a location."""
    with suppress(ValueError):
        hostname = urlparse(location).hostname
        if not hostname:
            return None

        return ip_address(hostname).version

    return None


def location_changed(ssdp_device: SsdpDevice, headers: CaseInsensitiveDict) -> bool:
    """Test if location changed for device."""
    new_location = headers.get_lower("location", "")
    if not new_location:
        return False

    # Device did not have any location, must be new.
    locations = ssdp_device.locations
    if not locations:
        return True

    if new_location in locations:
        return False

    # Ensure the new location is parsable.
    new_ip_version = ip_version_from_location(new_location)
    if new_ip_version is None:
        return False

    # We already established the location
    # was not seen before. If we have any location
    # saved that is the same ip version, we
    # consider the location changed
    return any(ip_version_from_location(location) == new_ip_version for location in locations)


class SsdpDeviceTracker:
    """
    Device tracker.

    Tracks `SsdpDevices` seen by the `SsdpListener`. Can be shared between `SsdpListeners`.
    """

    def __init__(self) -> None:
        """Initialize."""
        self.devices: dict[UniqueDeviceName, SsdpDevice] = {}
        self.next_valid_to: datetime | None = None

    def see_search(
        self, headers: CaseInsensitiveDict
    ) -> tuple[bool, SsdpDevice | None, DeviceOrServiceType | None, SsdpSource | None]:
        """See a device through a search."""
        if not valid_search_headers(headers):
            _LOGGER.debug("Received invalid search headers: %s", headers)
            return False, None, None, None

        udn = headers.get_lower("_udn")
        is_new_device = udn not in self.devices

        ssdp_device, new_location = self._see_device(headers)
        if not ssdp_device:
            return False, None, None, None

        search_target: SearchTarget = headers.get_lower("st")
        is_new_service = (
            search_target not in ssdp_device.advertisement_headers and search_target not in ssdp_device.search_headers
        )
        if is_new_service:
            _LOGGER.debug("See new service: %s, type: %s", ssdp_device, search_target)

        changed = (
            is_new_device
            or is_new_service
            or new_location
            or headers_differ_from_existing_search(ssdp_device, search_target, headers)
        )
        ssdp_source = SsdpSource.SEARCH_CHANGED if changed else SsdpSource.SEARCH_ALIVE

        # Update stored headers.
        search_headers = ssdp_device.search_headers
        if search_target in ssdp_device.search_headers:
            search_headers[search_target].replace(headers)
        else:
            search_headers[search_target] = headers

        return True, ssdp_device, search_target, ssdp_source

    def see_advertisement(
        self, headers: CaseInsensitiveDict
    ) -> tuple[bool, SsdpDevice | None, DeviceOrServiceType | None]:
        """See a device through an advertisement."""
        if not valid_advertisement_headers(headers):
            _LOGGER.debug("Received invalid advertisement headers: %s", headers)
            return False, None, None

        udn = headers.get_lower("_udn")
        is_new_device = udn not in self.devices

        ssdp_device, new_location = self._see_device(headers)
        if not ssdp_device:
            return False, None, None

        notification_type: NotificationType = headers.get_lower("nt")
        is_new_service = (
            notification_type not in ssdp_device.advertisement_headers
            and notification_type not in ssdp_device.search_headers
        )
        if is_new_service:
            _LOGGER.debug("See new service: %s, type: %s", ssdp_device, notification_type)

        notification_sub_type: NotificationSubType = headers.get_lower("nts")
        propagate = (
            notification_sub_type == NotificationSubType.SSDP_UPDATE
            or is_new_device
            or is_new_service
            or new_location
            or headers_differ_from_existing_advertisement(ssdp_device, notification_type, headers)
        )

        # Update stored headers.
        advertisement_headers = ssdp_device.advertisement_headers
        if notification_type in advertisement_headers:
            advertisement_headers[notification_type].replace(headers)
        else:
            advertisement_headers[notification_type] = headers

        return propagate, ssdp_device, notification_type

    def _see_device(self, headers: CaseInsensitiveDict) -> tuple[SsdpDevice | None, bool]:
        """See a device through a search or advertisement."""
        # Purge any old devices.
        now = headers.get_lower("_timestamp")
        self.purge_devices(now)
        if not (usn := headers.get_lower("usn")) or not (udn := udn_from_usn(usn)):
            # Ignore broken devices.
            return None, False

        valid_to = extract_valid_to(headers)

        if udn not in self.devices:
            # Create new device.
            ssdp_device = SsdpDevice(udn, valid_to)
            _LOGGER.debug("See new device: %s", ssdp_device)
            self.devices[udn] = ssdp_device
        else:
            ssdp_device = self.devices[udn]
            ssdp_device.valid_to = valid_to

        # Test if new location.
        new_location = location_changed(ssdp_device, headers)

        # Update device.
        ssdp_device.add_location(headers.get_lower("location"), valid_to)
        ssdp_device.last_seen = now
        if not self.next_valid_to or self.next_valid_to > ssdp_device.valid_to:
            self.next_valid_to = ssdp_device.valid_to

        return ssdp_device, new_location

    def unsee_advertisement(
        self, headers: CaseInsensitiveDict
    ) -> tuple[bool, SsdpDevice | None, DeviceOrServiceType | None]:
        """Remove a device through an advertisement."""
        if not valid_byebye_headers(headers):
            return False, None, None
        if (
            not (usn := headers.get_lower("usn"))
            or not (udn := udn_from_usn(usn))
            or not (ssdp_device := self.devices.get(udn))
        ):
            # Ignore broken devices and devices we don't know about.
            return False, None, None

        del self.devices[udn]
        # Update device before propagating it
        notification_type: NotificationType = headers.get_lower("nt")
        advertisement_headers = ssdp_device.advertisement_headers
        if notification_type in advertisement_headers:
            advertisement_headers[notification_type].replace(headers)
        else:
            advertisement_headers[notification_type] = CaseInsensitiveDict(headers)

        propagate = True  # Always true, if this is the 2nd unsee then device is already deleted.
        return propagate, ssdp_device, notification_type

    def get_device(self, headers: CaseInsensitiveDict) -> SsdpDevice | None:
        """Get a device from headers."""
        if not (usn := headers.get_lower("usn")) or not (udn := udn_from_usn(usn)):
            return None
        return self.devices.get(udn)

    def purge_devices(self, override_now: datetime | None = None) -> None:
        """Purge any devices for which the CACHE-CONTROL header is timed out."""
        now = override_now or datetime.now()
        if self.next_valid_to and self.next_valid_to > now:
            return
        self.next_valid_to = None
        to_remove = []
        for usn, device in self.devices.items():
            if now > device.valid_to:
                to_remove.append(usn)
            elif not self.next_valid_to or device.valid_to < self.next_valid_to:
                self.next_valid_to = device.valid_to
                device.purge_locations(now)
        for usn in to_remove:
            _LOGGER.debug("Purging device, USN: %s", usn)
            del self.devices[usn]


class SsdpListener:
    """SSDP Search and Advertisement listener."""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        async_callback: (
            Callable[[SsdpDevice, DeviceOrServiceType, SsdpSource], Coroutine[Any, Any, None]] | None
        ) = None,
        callback: (Callable[[SsdpDevice, DeviceOrServiceType, SsdpSource], None] | None) = None,
        source: AddressTupleVXType | None = None,
        target: AddressTupleVXType | None = None,
        loop: AbstractEventLoop | None = None,
        search_timeout: int = SSDP_MX,
        search_target: str = SSDP_ST_ALL,
        device_tracker: SsdpDeviceTracker | None = None,
    ) -> None:
        """Initialize."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        assert callback or async_callback, "Provide at least one callback"

        self.async_callback = async_callback
        self.callback = callback
        self.source, self.target = determine_source_target(source, target)
        self.loop = loop or asyncio.get_event_loop()
        self.search_timeout = search_timeout
        self.search_target = search_target
        self._device_tracker = device_tracker or SsdpDeviceTracker()
        self._advertisement_listener: SsdpAdvertisementListener | None = None
        self._search_listener: SsdpSearchListener | None = None

    async def async_start(self) -> None:
        """Start search listener/advertisement listener."""
        self._advertisement_listener = SsdpAdvertisementListener(
            on_alive=self._on_alive,
            on_update=self._on_update,
            on_byebye=self._on_byebye,
            source=self.source,
            target=self.target,
            loop=self.loop,
        )
        await self._advertisement_listener.async_start()

        self._search_listener = SsdpSearchListener(
            callback=self._on_search,
            loop=self.loop,
            source=self.source,
            target=self.target,
            timeout=self.search_timeout,
            search_target=self.search_target,
        )
        await self._search_listener.async_start()

    async def async_stop(self) -> None:
        """Stop scanner/listener."""
        if self._advertisement_listener:
            await self._advertisement_listener.async_stop()

        if self._search_listener:
            self._search_listener.async_stop()

    async def async_search(self, override_target: AddressTupleVXType | None = None) -> None:
        """Send a SSDP Search packet."""
        assert self._search_listener is not None, "Call async_start() first"
        self._search_listener.async_search(override_target)

    def _on_search(self, headers: CaseInsensitiveDict) -> None:
        """Search callback."""
        (
            propagate,
            ssdp_device,
            device_or_service_type,
            ssdp_source,
        ) = self._device_tracker.see_search(headers)

        if propagate and ssdp_device and device_or_service_type:
            assert ssdp_source is not None
            if self.async_callback:
                coro = self.async_callback(ssdp_device, device_or_service_type, ssdp_source)
                self.loop.create_task(coro)
            if self.callback:
                self.callback(ssdp_device, device_or_service_type, ssdp_source)

    def _on_alive(self, headers: CaseInsensitiveDict) -> None:
        """On alive."""
        (
            propagate,
            ssdp_device,
            device_or_service_type,
        ) = self._device_tracker.see_advertisement(headers)

        if propagate and ssdp_device and device_or_service_type:
            if self.async_callback:
                coro = self.async_callback(ssdp_device, device_or_service_type, SsdpSource.ADVERTISEMENT_ALIVE)
                self.loop.create_task(coro)
            if self.callback:
                self.callback(ssdp_device, device_or_service_type, SsdpSource.ADVERTISEMENT_ALIVE)

    def _on_byebye(self, headers: CaseInsensitiveDict) -> None:
        """On byebye."""
        (
            propagate,
            ssdp_device,
            device_or_service_type,
        ) = self._device_tracker.unsee_advertisement(headers)

        if propagate and ssdp_device and device_or_service_type:
            if self.async_callback:
                coro = self.async_callback(ssdp_device, device_or_service_type, SsdpSource.ADVERTISEMENT_BYEBYE)
                self.loop.create_task(coro)
            if self.callback:
                self.callback(ssdp_device, device_or_service_type, SsdpSource.ADVERTISEMENT_BYEBYE)

    def _on_update(self, headers: CaseInsensitiveDict) -> None:
        """On update."""
        (
            propagate,
            ssdp_device,
            device_or_service_type,
        ) = self._device_tracker.see_advertisement(headers)

        if propagate and ssdp_device and device_or_service_type:
            if self.async_callback:
                coro = self.async_callback(ssdp_device, device_or_service_type, SsdpSource.ADVERTISEMENT_UPDATE)
                self.loop.create_task(coro)
            if self.callback:
                self.callback(ssdp_device, device_or_service_type, SsdpSource.ADVERTISEMENT_UPDATE)

    @property
    def devices(self) -> Mapping[str, SsdpDevice]:
        """Get the known devices."""
        return self._device_tracker.devices



"""Unit tests for the DLNA DMS profile."""

import pytest

from async_upnp_client.client_factory import UpnpFactory
from async_upnp_client.const import HttpResponse
from async_upnp_client.exceptions import UpnpResponseError
from async_upnp_client.profiles.dlna import DmsDevice

from ..conftest import RESPONSE_MAP, UpnpTestNotifyServer, UpnpTestRequester, read_file


@pytest.mark.asyncio
async def test_async_browse_metadata() -> None:
    """Test retrieving object metadata."""
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dms:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = DmsDevice(device, event_handler=event_handler)

    # Object 0 is the root and must always exist
    requester.response_map[("POST", "http://dlna_dms:1234/upnp/control/ContentDir")] = HttpResponse(
        200,
        {},
        read_file("dlna/dms/action_Browse_metadata_0.xml"),
    )
    metadata = await profile.async_browse_metadata("0")
    assert metadata.parent_id == "-1"
    assert metadata.id == "0"
    assert metadata.title == "root"
    assert metadata.upnp_class == "object.container.storageFolder"
    assert metadata.child_count == "4"

    # Object 2 will give some different results
    requester.response_map[("POST", "http://dlna_dms:1234/upnp/control/ContentDir")] = HttpResponse(
        200,
        {},
        read_file("dlna/dms/action_Browse_metadata_2.xml"),
    )
    metadata = await profile.async_browse_metadata("2")
    assert metadata.parent_id == "0"
    assert metadata.id == "2"
    assert metadata.title == "Video"
    assert metadata.upnp_class == "object.container.storageFolder"
    assert metadata.child_count == "3"

    # Object that is an item and not a container
    requester.response_map[("POST", "http://dlna_dms:1234/upnp/control/ContentDir")] = HttpResponse(
        200,
        {},
        read_file("dlna/dms/action_Browse_metadata_item.xml"),
    )
    metadata = await profile.async_browse_metadata("1$6$35$1$1")
    assert metadata.parent_id == "1$6$35$1"
    assert metadata.id == "1$6$35$1$1"
    assert metadata.title == "Test song"
    assert metadata.upnp_class == "object.item.audioItem.musicTrack"
    assert metadata.artist == "Test artist"
    assert metadata.genre == "Rock & Roll"
    assert len(metadata.resources) == 1
    assert metadata.resources[0].uri == "http://dlna_dms:1234/media/2483.mp3"
    assert (
        metadata.resources[0].protocol_info == "http-get:*:audio/mpeg:DLNA.ORG_PN=MP3;DLNA.ORG_OP=01;DLNA.ORG_CI=0;"
        "DLNA.ORG_FLAGS=01700000000000000000000000000000"
    )
    assert metadata.resources[0].size == "2905191"
    assert metadata.resources[0].duration == "0:02:00.938"

    # Bad object ID should result in a UpnpError (HTTP 701: No such object)
    requester.exceptions.append(UpnpResponseError(status=701))
    with pytest.raises(UpnpResponseError) as err:
        await profile.async_browse_metadata("no object")

    assert err.value.status == 701


@pytest.mark.asyncio
async def test_async_browse_children() -> None:
    """Test retrieving children of a container."""
    # pylint: disable=too-many-statements
    requester = UpnpTestRequester(RESPONSE_MAP)
    factory = UpnpFactory(requester)
    device = await factory.async_create_device("http://dlna_dms:1234/device.xml")
    notify_server = UpnpTestNotifyServer(
        requester=requester,
        source=("192.168.1.2", 8090),
    )
    event_handler = notify_server.event_handler
    profile = DmsDevice(device, event_handler=event_handler)

    # Object 0 is the root and must always exist
    requester.response_map[("POST", "http://dlna_dms:1234/upnp/control/ContentDir")] = HttpResponse(
        200,
        {},
        read_file("dlna/dms/action_Browse_children_0.xml"),
    )
    result = await profile.async_browse_direct_children("0")
    assert result.number_returned == 4
    assert result.total_matches == 4
    assert result.update_id == 2333
    children = result.result
    assert len(children) == 4
    assert children[0].title == "Browse Folders"
    assert children[0].id == "64"
    assert children[0].child_count == "4"
    assert children[1].title == "Music"
    assert children[1].id == "1"
    assert children[1].child_count == "7"
    assert children[2].title == "Pictures"
    assert children[2].id == "3"
    assert children[2].child_count == "5"
    assert children[3].title == "Video"
    assert children[3].id == "2"
    assert children[3].child_count == "3"

    # Object 2 will give some different results
    requester.response_map[("POST", "http://dlna_dms:1234/upnp/control/ContentDir")] = HttpResponse(
        200,
        {},
        read_file("dlna/dms/action_Browse_children_2.xml"),
    )
    result = await profile.async_browse_direct_children("2")
    assert result.number_returned == 3
    assert result.total_matches == 3
    assert result.update_id == 2333
    children = result.result
    assert len(children) == 3
    assert children[0].title == "All Video"
    assert children[0].id == "2$8"
    assert children[0].child_count == "583"
    assert children[1].title == "Folders"
    assert children[1].id == "2$15"
    assert children[1].child_count == "2"
    assert children[2].title == "Recently Added"
    assert children[2].id == "2$FF0"
    assert children[2].child_count == "50"

    # Object that is an item and not a container
    requester.response_map[("POST", "http://dlna_dms:1234/upnp/control/ContentDir")] = HttpResponse(
        200,
        {},
        read_file("dlna/dms/action_Browse_children_item.xml"),
    )
    result = await profile.async_browse_direct_children("1$6$35$1$1")
    assert result.number_returned == 0
    assert result.total_matches == 0
    assert result.update_id == 2333
    assert result.result == []

    # Bad object ID should result in a UpnpError (HTTP 701: No such object)
    requester.exceptions.append(UpnpResponseError(status=701))
    with pytest.raises(UpnpResponseError) as err:
        await profile.async_browse_direct_children("no object")

    assert err.value.status == 701



"""async_upnp_client.search module."""

import asyncio
import logging
import socket
import sys
from asyncio import DatagramTransport
from asyncio.events import AbstractEventLoop
from ipaddress import IPv4Address, IPv6Address
from typing import Any, Callable, Coroutine, cast

from async_upnp_client.const import SsdpSource
from async_upnp_client.ssdp import (
    SSDP_DISCOVER,
    SSDP_MX,
    SSDP_ST_ALL,
    AddressTupleVXType,
    IPvXAddress,
    SsdpProtocol,
    build_ssdp_search_packet,
    determine_source_target,
    get_host_string,
    get_ssdp_socket,
)
from async_upnp_client.utils import CaseInsensitiveDict

_LOGGER = logging.getLogger(__name__)


class SsdpSearchListener:
    """SSDP Search (response) listener."""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        async_callback: (Callable[[CaseInsensitiveDict], Coroutine[Any, Any, None]] | None) = None,
        callback: Callable[[CaseInsensitiveDict], None] | None = None,
        loop: AbstractEventLoop | None = None,
        source: AddressTupleVXType | None = None,
        target: AddressTupleVXType | None = None,
        timeout: int = SSDP_MX,
        search_target: str = SSDP_ST_ALL,
        async_connect_callback: Callable[[], Coroutine[Any, Any, None]] | None = None,
        connect_callback: Callable[[], None] | None = None,
    ) -> None:
        """Init the ssdp listener class."""
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        assert callback is not None or async_callback is not None, "Provide at least one callback"

        self.async_callback = async_callback
        self.callback = callback
        self.async_connect_callback = async_connect_callback
        self.connect_callback = connect_callback
        self.search_target = search_target
        self.source, self.target = determine_source_target(source, target)
        self.timeout = timeout
        self.loop = loop or asyncio.get_event_loop()
        self._target_host: str | None = None
        self._transport: DatagramTransport | None = None

    def async_search(self, override_target: AddressTupleVXType | None = None) -> None:
        """Start an SSDP search."""
        assert self._transport is not None
        sock: socket.socket | None = self._transport.get_extra_info("socket")
        _LOGGER.debug(
            "Sending SEARCH packet, transport: %s, socket: %s, override_target: %s",
            self._transport,
            sock,
            override_target,
        )

        assert self._target_host is not None, "Call async_start() first"
        packet = build_ssdp_search_packet(self.target, self.timeout, self.search_target)

        protocol = cast(SsdpProtocol, self._transport.get_protocol())
        target = override_target or self.target
        protocol.send_ssdp_packet(packet, target)

    def _on_data(self, request_line: str, headers: CaseInsensitiveDict) -> None:
        """Handle data."""
        if headers.get_lower("man") == SSDP_DISCOVER:
            # Ignore discover packets.
            return
        if headers.get_lower("nts"):
            _LOGGER.debug("Got non-search response packet: %s, %s", request_line, headers)
            return

        if _LOGGER.isEnabledFor(logging.DEBUG):
            _LOGGER.debug(
                "Received search response, _remote_addr: %s, USN: %s, location: %s",
                headers.get_lower("_remote_addr", ""),
                headers.get_lower("usn", "<no USN>"),
                headers.get_lower("location", ""),
            )
        headers["_source"] = SsdpSource.SEARCH
        if self._target_host and self._target_host != headers["_host"]:
            return
        if self.async_callback:
            coro = self.async_callback(headers)
            self.loop.create_task(coro)
        if self.callback:
            self.callback(headers)

    def _on_connect(self, transport: DatagramTransport) -> None:
        sock: socket.socket | None = transport.get_extra_info("socket")
        _LOGGER.debug("On connect, transport: %s, socket: %s", transport, sock)
        self._transport = transport
        if self.async_connect_callback:
            coro = self.async_connect_callback()
            self.loop.create_task(coro)
        if self.connect_callback:
            self.connect_callback()

    @property
    def target_ip(self) -> IPvXAddress:
        """Get target IP."""
        if len(self.target) == 4:
            return IPv6Address(self.target[0])

        return IPv4Address(self.target[0])

    async def async_start(self) -> None:
        """Start the listener."""
        _LOGGER.debug("Start listening for search responses")

        sock, _source, _target = get_ssdp_socket(self.source, self.target)
        if sys.platform.startswith("win32"):
            address = self.source
            _LOGGER.debug("Binding socket, socket: %s, address: %s", sock, address)
            sock.bind(address)

        if not self.target_ip.is_multicast:
            self._target_host = get_host_string(self.target)
        else:
            self._target_host = ""

        loop = self.loop
        await loop.create_datagram_endpoint(
            lambda: SsdpProtocol(
                loop,
                on_connect=self._on_connect,
                on_data=self._on_data,
            ),
            sock=sock,
        )

    def async_stop(self) -> None:
        """Stop the listener."""
        if self._transport:
            self._transport.close()


async def async_search(
    async_callback: Callable[[CaseInsensitiveDict], Coroutine[Any, Any, None]],
    timeout: int = SSDP_MX,
    search_target: str = SSDP_ST_ALL,
    source: AddressTupleVXType | None = None,
    target: AddressTupleVXType | None = None,
    loop: AbstractEventLoop | None = None,
) -> None:
    """Discover devices via SSDP."""
    # pylint: disable=too-many-arguments,too-many-positional-arguments
    loop_: AbstractEventLoop = loop or asyncio.get_event_loop()
    listener: SsdpSearchListener | None = None

    async def _async_connected() -> None:
        nonlocal listener
        assert listener is not None
        listener.async_search()

    listener = SsdpSearchListener(
        async_callback=async_callback,
        loop=loop_,
        source=source,
        target=target,
        timeout=timeout,
        search_target=search_target,
        async_connect_callback=_async_connected,
    )

    await listener.async_start()

    # Wait for devices to respond.
    await asyncio.sleep(timeout)

    listener.async_stop()
